"""Classes for completions

The idea is inspired from
https://github.com/pallets/click/pull/1622

Some of the code is borrowing there, under following LICENSE:

Copyright 2014 Pallets

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
Neither the name of the copyright holder nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import os
import sys
import re
from typing import Optional, Iterator, List, Callable, Type, Union, Tuple
from pathlib import Path
from hashlib import sha256

COMPLETION_SCRIPT_BASH = """
%(complete_func)s() {
    local IFS=$'\\n'
    local response
    response=$( env COMP_WORDS="${COMP_WORDS[*]}" \\
                COMP_CWORD=$COMP_CWORD \\
                %(complete_shell_var)s=bash %(complete_script)s )
    for completion in $response; do
        IFS=$'\\t' read value type <<< "$completion"
        if [[ $type == 'dir' ]]; then
            COMREPLY=()
            compopt -o dirnames
        elif [[ $type == 'file' ]]; then
            COMREPLY=()
            compopt -o filenames
        elif [[ $type == 'plain' ]]; then
            COMPREPLY+=($value)
        fi
    done
     return 0
}
complete -o default -F %(complete_func)s %(script_name)s
"""

COMPLETION_SCRIPT_ZSH = """
#compdef %(script_name)s
%(complete_func)s() {
    local -a completions
    local -a completions_with_descriptions
    local -a response
    (( ! $+commands[%(script_name)s] )) && return 1
    response=("${(@f)$( env COMP_WORDS=\"${words[*]}\" \\
                        COMP_CWORD=$((CURRENT-1)) \\
                        %(complete_shell_var)s=\"zsh\" \\
                        %(complete_script)s )}")
    for key type descr in ${response}; do
      if [[ "$type" == "plain" ]]; then
        if [[ "$descr" == "" ]]; then
          completions+=("$key")
        else
          completions_with_descriptions+=("$key":"$descr")
        fi
      elif [[ "$type" == "dir" ]]; then
        _path_files -/
      elif [[ "$type" == "file" ]]; then
        _path_files -f
      fi
    done
     if [ -n "$completions_with_descriptions" ]; then
        _describe -V unsorted completions_with_descriptions -U
    fi
     if [ -n "$completions" ]; then
        compadd -U -V unsorted -a completions
    fi
    compstate[insert]="menucomplete"
}
compdef %(complete_func)s %(script_name)s
"""

COMPLETION_SCRIPT_FISH = """
function %(complete_func)s_complete;
    set -l response;
    set -lx COMP_WORDS (commandline -op)
    set -lx COMP_CWORD ( \\
        math (contains -i -- (commandline -t) $COMP_WORDS; or echo 0)-1 \\
    )
    set -lx %(complete_shell_var)s fish
    eval %(complete_script)s | while read completion
        echo $completion | read -d (echo -e "\\t") -l -a metadata
        if [ "$metadata[2]" = "dir" ];
            __fish_complete_directories $metadata[1] | sed "s#^#$metadata[3]#";
        else if [ "$metadata[2]" = "file" ];
            __fish_complete_path $metadata[1] | sed "s#^#$metadata[3]#";
        else if [ "$metadata[2]" = "plain" ];
            echo -n $metadata[1];
            echo -ne "\\t";
            echo $metadata[3];
        end;
    end;
end;

# Don't do complete until <prog> is hit
function %(complete_func)s_condition;
    set -l COMP_WORDS (commandline -op)
    set -l comp_script %(complete_script)s
    set -l len_words (count $COMP_WORDS)
    set -l len_script (count $comp_script)
    set -l incomplete (commandline -t)
    if [ $len_script -eq 1 ];
        return 0;
    end
    # we haven't hit the script or module
    # go ahead do the complete only when
    # len_words > len_script
    # if len_words == len_script, then requires incomplete == ""
    if [ $len_words -lt $len_script ]
        return 1
    else if [ $len_words -eq $len_script -a -n "$incomplete" ]
        return 1
    else if [ $len_script -eq 2 ]
        [ "$COMP_WORDS[2]" = "$comp_script[2]" ]; and return 0; or return 1
    else if [ $len_script -eq 3 ]
        [ "$COMP_WORDS[2]" = "-m" -a "$comp_script[3]" = "$COMP_WORDS[3]" ]; \\
            and return 0; or return 1
    end
    return 1
end;

complete --no-files --command %(script_name)s \\
    --condition "%(complete_func)s_condition" \\
    --arguments "(%(complete_func)s_complete)"
"""


def split_arg_string(string: str) -> List[str]:
    # pylint: disable=line-too-long
    """Given an argument string this attempts to split it into small parts.

    Borrowed from
    https://github.com/pallets/click/blob/3984f9efce5a0d15f058e1abe1ea808c6abd243a/src/click/parser.py#L106

    Args:
        string: The string to be split

    Returns:
        List of split pieces
    """
    # pylint: enable=line-too-long
    ret: List[str] = []
    for match in re.finditer(
            r"('([^'\\]*(?:\\.[^'\\]*)*)'|"
            r"\"([^\"\\]*(?:\\.[^\"\\]*)*)\"|\S+)\s*",
            string,
            re.S,
    ):
        arg = match.group().strip()
        if arg[:1] == arg[-1:] and arg[:1] in "\"'":
            arg = arg[1:-1].encode(
                "ascii", "backslashreplace"
            ).decode("unicode-escape")
        try:
            arg = type(string)(arg)
        except UnicodeError: # pragma: no cover
            pass
        ret.append(arg)
    return ret


class Completer:
    """Main completion handler

    Attributes:
        comp_shell: The shell where the completion will be conducted
            One of ['', 'bash', 'fish', 'zsh']
            Obtained from environment
        comp_words: The words have been entered before completion
        comp_curr: The current word for completion
        comp_prev: The previous word matched
    """

    def __init__(self):
        """Constructor

        Read the environment variables
        """
        complete_prepared = self._prepare_complete()

        self.comp_shell: str = complete_prepared[0]
        self.comp_words: Optional[List[str]] = (
            complete_prepared[1] if self.comp_shell else None
        )
        self.comp_curr: Optional[str] = (
            complete_prepared[2] if self.comp_shell else None
        )
        self.comp_prev: Optional[str] = (
            None if not self.comp_shell or not self.comp_words
            else self.comp_words[-1]
        )

    @property
    def progvar(self) -> str:
        """Get the program name that can be used as a variable"""
        return re.sub(r'[^\w_]+', '', self.prog)

    @property
    def uid(self) -> str:
        """Get the uid based on the raw program name

        This is used as the prefix or suffix of some shell function names
        """
        return sha256(self.prog.encode()).hexdigest()[:6]

    def _prepare_complete(self) -> Tuple[
            str, Optional[List[str]], Optional[str]
    ]:
        """Prepare for completion, get the env variables"""
        env_name: str = f"{self.progvar}_COMPLETE_SHELL_{self.uid}".upper()
        shell: str = os.environ.get(env_name, '')
        if not shell:
            return shell, None, ''

        comp_words: List[str] = split_arg_string(os.environ['COMP_WORDS'])
        comp_cword: int = int(os.environ['COMP_CWORD'] or 0)

        current: str = ''
        if comp_cword >= 0:
            try:
                current = comp_words[comp_cword]
            except IndexError:
                pass

        has_python: bool = 'python' in Path(comp_words[0]).stem
        if has_python and len(comp_words) == 1:
            sys.exit(0)

        is_module: bool = has_python and comp_words[1] == '-m'
        if is_module and (len(comp_words) < 3 or comp_words[2] != self.prog):
            sys.exit(0)

        if has_python and not is_module and comp_words[1] != self.prog:
            sys.exit(0)

        comp_words = comp_words[
            (3 if is_module else 2 if has_python else 1):
        ]

        if current and comp_words and comp_words[-1] == current:
            comp_words.pop(-1)

        if shell == 'bash' and comp_words:
            # bash splits '--choice=' to ['--choice'] and '=', and
            # '--choice=l' to ['--choice', '='] and 'l'
            # We can't distinguish if user really enters '--choice=' or
            # '--choice =', but this is the best way to implement this.
            # Also, bash doesn't replace the current,
            # so we just need to get the unfinished part
            if current == '=':
                current = '' # force the unfinished part
            elif current and comp_words[-1] == '=' and len(comp_words) > 1:
                # pop out the '=' so to force th unfinished part
                comp_words.pop()

        return shell, comp_words, current

    def _post_complete(
            self,
            completions: Iterator[Tuple[str, str, str]]
    ) -> Optional[str]:
        """Post processing the completions

        Filter only completions with given current word/prefix
        If non-fish, don't give the description
        """
        for comp in completions:
            if self.comp_shell == 'fish':
                yield '\t'.join(comp)
            elif self.comp_shell == 'zsh':
                yield '\n'.join((comp[0] or ' ', comp[1], comp[2]))
            else:
                yield '\t'.join((comp[0] or ' ', comp[1]))

    def shellcode(self,
                  shell: str,
                  python: Optional[str] = None,
                  module: bool = False) -> None:
        """Generate the shell code to be integrated

        For bash, it should be appended to ~/.profile
        For zsh, it should be appended to ~/.zprofile
        For fish, it should be appended to
            ~/.config/fish/completions/{prog}.fish
            If python is provided, this should go to `python.fish` rather than
            the `{prog}.fish`

        Args:
            shell: The shell to generate the code for.
            python: The python name or path to invoke completion.
            module: Whether do completion for `python -m <prog>`

        Raises:
            ValueError: if shell is not one of bash, zsh and fish
        """

        if shell == 'zsh':
            return self._shellcode_zsh(python=python, module=module)
        if shell == 'fish':
            return self._shellcode_fish(python=python, module=module)
        if shell == 'bash':
            return self._shellcode_bash(python=python, module=module)
        raise ValueError(f'Shell not supported: {shell}')

    def _shellcode_bash(self, python: Optional[str], module: bool) -> str:
        # type: (Optional[str]) -> str
        """Generate the shell code for bash"""
        complete_shell_var: str = (
            f"{self.progvar}_COMPLETE_SHELL_{self.uid}"
        ).upper()
        complete_script: str = (
            "$1" if not python
            else f"$1 {self.prog}" if not module
            else f"$1 -m {self.prog}"
        )
        complete_func: str = f"_{self.progvar}_completion_{self.uid}"
        return COMPLETION_SCRIPT_BASH % dict(
            complete_func=complete_func,
            complete_shell_var=complete_shell_var,
            complete_script=complete_script,
            script_name=python or self.prog
        )

    def _shellcode_fish(self, python: Optional[str], module: bool) -> str:
        # type: (Optional[str]) -> str
        """Generate the shell code for fish"""
        complete_shell_var: str = (
            f"{self.progvar}_COMPLETE_SHELL_{self.uid}"
        ).upper()
        complete_script: str = (
            "$COMP_WORDS[1]" if not python
            else f"$COMP_WORDS[1] {self.prog}" if not module
            else f"$COMP_WORDS[1] -m {self.prog}"
        )
        complete_func: str = f"__fish_{self.progvar}_{self.uid}"
        return COMPLETION_SCRIPT_FISH % dict(
            complete_func=complete_func,
            complete_shell_var=complete_shell_var,
            complete_script=complete_script,
            script_name=python or self.prog
        )

    def _shellcode_zsh(self, python: Optional[str], module: str) -> str:
        # type: (Optional[str]) -> str
        """Generate the shell code for zsh"""
        complete_shell_var: str = (
            f"{self.progvar}_COMPLETE_SHELL_{self.uid}"
        ).upper()
        complete_script: str = (
            "$words[1]" if not python
            else f"$words[1] {self.prog}" if not module
            else f"$words[1] -m {self.prog}"
        )
        complete_func: str = f"_{self.progvar}_completion_{self.uid}"
        return COMPLETION_SCRIPT_ZSH % dict(
            complete_func=complete_func,
            complete_shell_var=complete_shell_var,
            complete_script=complete_script,
            script_name=python or self.prog
        )

    def _get_param_by_prefixed(self, prefixed: str) -> Optional[Type['Param']]:
        """Get the parameter by the given prefixed name"""
        for param in self._all_params(True):
            if any(param._prefix_name(name) == prefixed
                   for name in param.names):
                return param
        return None

    def _parse_completed(self) -> Tuple[
            Optional[List[Type['Param']]], Optional[bool],
            Optional[str], Optional[List[str]]
    ]:
        """Parse completed parameters/commands, and give
        the rest unmatched words. If command matched, also return the command

        Returns:
            A tuple of:
                - A list of completed parameters.
                - A boolean value indicating whether all required parameters
                    has been completed
                - Command name if a command matched
                - Rest of words after the command is matched
        """
        for i, word in enumerate(self.comp_words):
            if word in self.commands:
                return None, None, word, self.comp_words[i+1:]

        unmatched_required: bool = False
        matched: List[Type['Param']] = []
        matched_append: Callable = matched.append
        for param in self._all_params(True):
            if any(param._prefix_name(name) in self.comp_words
                   for name in param.names):
                matched_append(param)
            elif param.required:
                unmatched_required = True

        return matched, not unmatched_required, None, None

    def complete(self) -> Iterator[str]:
        """Yields the completions

        Yields:
            The strings as completion candidates
        """
        yield from self._post_complete(self._complete())

    def _complete(self) -> Iterator[Tuple[str, str, str]]:
        """Provide the completion candidates

        The rules are:
        1. When there are uncompleted required parameters, don't give command
           candidates
        2. Otherwise, give both command and parameter candidates
        """
        (completed, all_required_completed,
         command, rest) = self._parse_completed()

        if command:
            self.commands[command].comp_shell = self.comp_shell
            self.commands[command].comp_words = rest
            self.commands[command].comp_curr = self.comp_curr
            self.commands[command].comp_prev = rest[-1] if rest else None
            # make sure that help parameters or commands are added
            self.commands[command].parse()
            # sys.exit(0)

        completions: Optional[Union[
            str,
            Iterator[Tuple[str]],
            Iterator[Tuple[str, str]],
            Iterator[Tuple[str, str, str]]
        ]] = ''
        param: Optional[Type['Param']] = None
        # see if comp_curr is something like '--arg=x'
        if self.comp_curr and '=' in self.comp_curr:
            prefixed, val = self.comp_curr.split('=', 1)
            param = self._get_param_by_prefixed(prefixed)
            completions = param.complete_value(
                current=val, prefix=f'{prefixed}='
            ) if param else completions
        else:
            param = self._get_param_by_prefixed(self.comp_prev)
            completions = param.complete_value(
                current=self.comp_curr
            ) if param else completions

        if param:
            if completions is None:
                return # StopIteration
            if completions is not None and completions:
                for completion in completions:
                    yield ((completion[0], 'plain', '')
                           if len(completion) == 1
                           else (completion[0], 'plain', completion[1])
                           if len(completion) == 2
                           else completion)
                return # StopIteration, dont go further

        # no param or completions == ''
        for param in self._all_params(True):
            if param.type == 'ns':
                continue
            if param in completed and not param.complete_relapse:
                continue

            for prefixed_name, desc in param.complete_name(self.comp_curr):
                yield (prefixed_name, 'plain', desc)

        if all_required_completed:
            # see if we have any commands
            for command_name, command in self.commands.items():
                if command_name.startswith(self.comp_curr):
                    yield (command_name, "plain",
                           "Command: " + command.desc[0].splitlines()[0])


class CompleterParam:
    """Class for a parameter dealing with completion"""
    complete_relapse: bool = False

    def complete_value(
            self,
            current: str,
            prefix: str = ''
    ) -> Optional[Union[str, Iterator[Tuple[str]], Iterator[Tuple[str, str]],
                        Iterator[Tuple[str, str, str]]]]:
        """Give the completion candidates

        Args:
            current: Current prefix

        Returns:
            None when there are no candidates, nor should we have next
                paramters/commands as candidates (requiring a value).
                An empty string if we should put next parameters/commands
                as candidates. Otherwise yields
                The candidates should be either 1-, 2-, or 3-element tuple.
                If 1-element, type plain and no description implied.
                If 2-element, type plain and 2nd element should be description.
                If 3-element, 2nd element the type, 3rd the description.
        """
        if callable(self.complete_callback):
            return self.complete_callback(current, prefix)
        return None

    def complete_name(self, current: str) -> Iterator[Tuple[str, str]]:
        """Give the completion name candidates

        Args:
            current: The current prefix or word under cursor

        Returns:
            An iterator of a tuple including the prefixed name and description.
        """
        for name in self.names:
            prefixed: str = self._prefix_name(name)
            if prefixed.startswith(current):
                yield (prefixed, self.desc[0].splitlines()[0])



"""For users to test arbitrary parsing"""
import sys
import logging
from pathlib import Path
from rich import print # pylint: disable=redefined-builtin
from . import defaults
from . import Params, Namespace
from .utils import logger

# pylint: disable=invalid-name
defaults.CONSOLE_WIDTH = 100

params = Params(prog='pyparam', desc="""\
An exhibition showing all supported types of parameters and \
some features by running {prog}.

```python
# We can also insert code block in the description.
print('Hello pyparam!')
```

>>> # This is another example of code block using python console
>>> print('Hello pyparam!')""")

params.add_param('d,debug', False,
                 desc='Show the debug logging of the parsing process?')

predefined = params.add_command([
    "p", "pred", "pd", "pdf", "predef", "predefined", "pre-defined",
    "pre-defined-args"
], desc="Some predefined arguments.", usage="""\
{prog} -i-1 --in . --py 1
{prog} --int=-1 --in . --py "(1,2,3)" -f0.1 -ccc
{prog} --int=-1 --in . --py "[1,2,3]" --float 0.1 --count 2
{prog} --int=-1 --in . --py True -b true --float 0.1
{prog} -i0 -b1 --choice=large
{prog} --int=-1 --bool false --float 0.1 --auto \
"{{\\\"a\\\": 1}}" --py "{{1, 2, 3}}" """)

predefined.add_param('i, int', required=True, type=int, desc=[
    "An argument whose value will be casted into an integer.",
    "You can also try the short name with attached value: `-i1`.",
    "Negative values should be passed using `=`: ",
    "`-i-1` or `--int=-1`",
    "Same for other values starting with `-`. Default: 0"
])
predefined.add_param('f, float', type_frozen=False, type=float, desc=[
    "An argument whose value will be casted into float.",
    "You can also try `--float=1e-3`",
    "You can even overwrite the type of the argument from command line by "
    "`--float:int=1`"
])
predefined.add_param('b, bool', type=bool, desc=[
    "A boolean/flag argument. ",
    "If it is hit by itself, `True` will be used. However, it can consume "
    "one of following values: [true, TRUE, True, 1, false, FALSE, False, 0]"
])
predefined.add_param('c, count', max=3, type='count', desc=[
    "A count argument, you can do -c, -cc, -ccc, etc.",
    "It must have a short name (`c` here for example). ",
    "You can also use it like an integer argument with long name. "
    "For example: --count 3",
    "The default value has to be `0` if specified. A max value can also "
    "be defined while adding this argument."
])
predefined.add_param('auto', desc=[
    "An argument whose value will be automatically casted.",
    "```python",
    "# Value received => Value/Type casted into",
    "'True', 'TRUE', 'true' => True",
    "'False', 'FALSE', 'false' => False",
    "'1', '2', '-1' => int",
    "'1.1', '2.0', '-1.0' => float",
    "'{{\"a\": 1}}' => {{\"a\": 1}} # json",
    "```",
    "If you don't want the value to be casted. Declare the argument with "
    "type `str`"
])
predefined.add_param('in', type='path', required=True, desc=[
    "An argument whose value will be casted into `pathlib.Path`.",
    "Since the argument name is a python keyword, you may not be able to "
    "fetch the value by:",
    "```python",
    "parsed = params.parse()",
    "parsed.in",
    "# you can do parsed['in'] instead.",
    "```",
    "You can use a callback here to check if the path exists:",
    ">>> def callback(path):",
    ">>>    if not path.exists():",
    ">>>        # You can also return the error using lambda",
    ">>>        raise ValueError('Path does not exist.')",
    ">>>    return path"
], callback=(lambda path: ValueError("Path does not exist.")
             if not path.exists() else path))
predefined.add_param("py", required=True, type="py", desc=[
    "Value will be evaluated using `ast.literal_eval`"
])
predefined.add_param("str", default='size', type="str", desc=[
    "Value will be kept anyway as a string."
])
predefined.add_param("json", type="json", default="{}", desc=[
    "Value will be converted using `json.loads`"
])
predefined.add_param("choice", type="choice", desc=[
    "One of the choices: small, medium and large.",
    "Callback can also be apply to modify the value:",
    ">>> def callback(value, all_values):",
    ">>>    # Use the value of argument '--str'",
    ">>>    return f'{{value}} {{all_values.str}}'"
], choices=[
    "small", "medium", "large"
], default="medium", callback=(lambda value, all_values:
                               f"{value} {all_values.str}"))
predefined.add_param("list", default=[1, 2, 3], type=list, desc=[
    "List/Array argument.",
    "You can pass the values one by one like this: `--list 1 2 3`.",
    "Or like this: --list 1 --list 2 --list 3",
    "List argument is incremental, meaning the values will be added to "
    "the values. To stop doing that, you can direct users to reset it using: ",
    "`--list:reset 4 5 6` or ",
    "`--list:reset 4 --list 5 --list 6`",
    "You can also set a subtype for list elements, including the scalar types. "
    "For exapmle: ",
    "`--list:list:bool 0 1 0` will produce [False, True, False]."
])

nsparams = params.add_command(
    'n, nsparams',
    help_on_void=False,
    desc=[
        "Namespace parameters. ",
        "No required arguments under this command, "
        "`help_on_void` is set to False.",
        "To bring up the help page, try arguments `-h/--help/-H`"
    ])
nsparams.add_param("config.nlayers", type='int', default=8, desc=[
    "Number of layers."
])
nsparams.add_param("config.num-heads", type='int', default=8, desc=[
    "Number of heads."
])
# give some description to put it on help page
nsparams.add_param(
    "config.nested, config.nt",
    type='ns',
    desc="""\
Nested configurations under config.
You can also apply callback to change the value of the whole namespace:
```python
def callback(value, all_values):
    value.gpus += all_values.config.nlayers
    return value
```""",
    callback=(
        lambda value, all_values: setattr(
            value, 'gpus', value.gpus + all_values.config.nlayers
        ) or value
    )
)
nsparams.add_param("config.nested.gpus", type='int', default=1, desc=[
    "Number of GPUs to use."
])

nested_cmd = params.add_command('nested-cmd', desc="Nested commands")
nested_cmd2 = nested_cmd.add_command('nested-cmd2',
                                     desc="A command of `nested-cmd`")



params.add_command("a, arbi, arbitrary",
                   desc="No predefined arguments, but you can "
                   "pass arbitrary arguments and see how "
                   "`pyparam` parses them",
                   arbitrary=True)

fromfile = params.add_command("f, fromfile", desc="""\
Load parameter definition from file.
You can load parameter definitions from any file type that \
`python-simpleconf` supports.
For example, a toml file:

```toml
[params.arg]
desc = "An argument"
```

See `example.toml` in the repo.
""")
fromfile.from_arg("file", desc="The file to load parameters from.")

complete = params.add_command("complete", desc="""\
Generate shell code for completions.

For bash:
```sh
$ python -m pyparam complete --shell bash --module >> ~/.profile
```

For zsh:
```sh
$ python -m pyparam complete --shell zsh --module >> ~/.zprofile
```

For fish:
```sh
$ python -m pyparam complete --shell fish --module \\
    >> ~/.config/fish/completions/python.fish
```
""")
complete.add_param('shell', required=True, type='choice',
                   choices=['bash', 'fish', 'zsh'],
                   desc='The shell where the program will be running. '
                   'One of {choices}.')
complete.add_param('script', default=False,
                   desc='Generate shell code for `python <prog>`')
complete.add_param('module', default=False,
                   desc='Generate shell code for `python -m <prog>`')
complete.add_param('py,python', default=Path(sys.executable).name,
                   desc='The python executable.')
complete.add_param('dir', type='dir', default=Path(__file__).parent,
                   desc='This does nothing but just show you '
                   'how directory completion works.')
complete.add_param('choice', type='choice',
                   choices=['xsmall', 'small', 'medium', 'large', 'xlarge'],
                   desc='This does nothing but just show you '
                   'how complete callback works.',
                   complete_callback=None)

def vars_ns(ns, depth=None):
    """Get the vars of a namespace"""
    ret = vars(ns)
    for key, val in ret.items():
        if (depth is None or depth > 0) and isinstance(val, Namespace):
            ret[key] = vars_ns(val, None if depth is None else depth - 1)
    return ret

def main():
    """Main entry"""
    parsed = params.parse()

    if parsed.debug:
        logger.setLevel(logging.DEBUG)
        params.parse()

    if parsed.__command__ == 'complete':
        print(params.shellcode(
            shell=parsed.complete.shell,
            python=parsed.complete.python
            if parsed.complete.script or parsed.complete.module else None,
            module=parsed.complete.module
        ))
    else:
        print()
        print("Arguments passed in:")
        print()
        print(vars_ns(parsed))

if __name__ == "__main__":
    main()



"""Defaults for pyparam

Holds some default values for pyparam.
To change any of the CONSOLE_WIDTH, HELP_SECTION_INDENT or HELP_OPTION_WIDTH:
```python
from pyparam import defaults
default.CONSOLE_WIDTH = 100
```

Attributes:
    POSITIONAL: The name of positional parameter
    TYPE_NAMES: The type name mappings to get the type name from aliases
        Do not modify this variable. It is maintained by
        `pyparam.param.regiest_param`
    CONSOLE_WIDTH: The total width for the help page.
    HELP_SECTION_INDENT: The indentation for the contents in a section
    HELP_OPTION_WIDTH: The width that the option name and type take up in
        the help page.
"""
from typing import Dict, Optional
from diot import Diot

POSITIONAL: str = ''

# - Single value types:
#     auto, int, str, float, bool, count, py, json
# - Complex value types:
#     list[<single/complex value type>], ns
TYPE_NAMES: Dict[str, str] = {}

CONSOLE_WIDTH: Optional[int] = 80
# indention for the contents of each section
HELP_SECTION_INDENT: int = 2
# The width of the options in help
HELP_OPTION_WIDTH: int = 34

# Default attribute values for a Params object
# This, as well as default attribute values for Param object,
# are useful to reduce the size of a dumped file
PARAMS: Diot = Diot(
    desc=['Not described.'],
    help_keys=['h', 'help'],
    help_cmds=['help'],
    help_on_void=True,
    prefix='auto',
    theme='default',
    usage=None,
    arbitrary=False,
)

PARAM: Diot = Diot(
    type=None,
    desc=['Not described.'],
    default=None,
    show=True,
    type_frozen=True,
    argname_shorten=True,
    required=False
)



"""Definition of a single parameter

Attributes:
    PARAM_MAPPINGS: The type to Param mappings, used for params to init a
        parameter. Managed by `register_param`
"""
# pylint: disable=too-many-lines
import ast
import json
from typing import (Optional, List, Any, Callable, Type,
                    Dict, Set, Tuple, Union, Iterator)
from itertools import product
from pathlib import Path
from diot import OrderedDiot
from .utils import cast_to, parse_type, logger, TYPE_NAMES, Namespace
from .defaults import POSITIONAL
from .completer import CompleterParam
from .exceptions import (PyParamException, PyParamValueError,
                         PyParamTypeError, PyParamNameError)

PARAM_MAPPINGS: Dict[str, Type['Param']] = {}

class Param(CompleterParam):
    """Base class for parameter

    Args:
        names: The names of the parameter
        default: The default value
        desc: The description of the parameter
        prefix: The prefix of the parameter on the command line
        show: Whether this parameter should show on help page
        required: Whether this parameter is required
        subtype: The subtype of the parameter if
            this is a complex type
        type_frozen: Whether the type is frozen
            (not allowing overwritting from command line)
        callback: The callback to modify the final value
        argname_shorten: Whether show shortened name for parameters
            under namespace parameters
        complete_callback: The callback for complete the values of the
            parameter
        **kwargs: Additional keyword arguments

    Attributes:
        names: The names of the parameter
        default: The default value
        prefix: The prefix of the parameter on the command line
        show: Whether this parameter should show on help page
        required: Whether this parameter is required
        subtype: The subtype of the parameter if this is a complex type
        type_frozen: Whether the type is frozen
            (not allowing overwritting from command line)
        callback: The callback to modify the final value
        argname_shorten: Whether show shortened name for the parameters
            under namespace parameters
        hit: Whether the parameter is just hit
        ns_param: The namespace parameter where this parameter
            is under
        complete_callback: The callback for complete the values of the
            parameter
        is_help: Whether this is a help parameter
        _desc: The raw description of the parameter
        _stack: The stack to push the values
        _value_cached: The cached value calculated from the stack
        _kwargs: other kwargs
    """
    # pylint: disable=too-many-instance-attributes
    type: Optional[str] = None
    type_aliases: List[str] = []

    @classmethod
    def on_register(cls):
        """Opens opportunity to do something when a parameter is registered"""

    def __init__(self,
                 names: List[str],
                 default: Any,
                 desc: Optional[List[str]],
                 prefix: str = 'auto',
                 show: bool = True,
                 required: bool = False,
                 subtype: Optional[bool] = None,
                 type_frozen: bool = True,
                 callback: Optional[Callable] = None,
                 complete_callback: Optional[Callable] = None,
                 argname_shorten: bool = True,
                 **kwargs: Dict[str, Any]):
        # pylint: disable=too-many-arguments
        """Constructor"""
        self.names: List[str] = names
        self.default: Any = default
        self.prefix: str = prefix
        self.show: bool = show
        self.required: bool = required
        self.subtype: Optional[str] = subtype
        self.type_frozen: bool = type_frozen
        self.callback: Optional[Callable] = callback
        self.complete_callback: Optional[Callable] = complete_callback
        self.argname_shorten: bool = argname_shorten
        self.hit: bool = False
        self.is_help: bool = False
        self.ns_param: Optional["ParamNamespace"] = None
        self._desc: List[str] = desc or ['No description.']
        self._stack: List[Any] = []
        self._value_cached: Optional[Any] = None
        self._kwargs: Dict[str, Any] = kwargs

        # check if I am under a namespace
        # Type: List[List[str]], List[str]
        self._namespaces, self.terminals = self._extract_namespaces()

    def _extract_namespaces(self) -> Tuple[List[List[str]], List[str]]:
        """Extract the namespace and terminal names"""
        nparts: Set[int] = set()
        namespaces: List[Set[str]] = []
        terminals: Set[str] = set()
        for name in self.names:
            parts: List[str] = name.split('.')
            nparts.add(len(parts))
            if len(nparts) > 1:
                raise PyParamNameError(
                    "Parameter names must have the same number of namespaces."
                )
            namespaces = namespaces or [{part} for part in parts[:-1]]
            for i, part in enumerate(parts[:-1]):
                namespaces[i].add(part)
            terminals.add(parts[-1])
        return [list(ns) for ns in namespaces], list(terminals)

    def _prefix_name(self, name: str) -> str:
        """Add prefix to a name

        Args:
            name: Name to add prefix to

        Returns:
            Name with prefix added
        """
        name_to_check: str = name.split('.', 1)[0]
        if self.prefix == 'auto':
            return f"-{name}" if len(name_to_check) <= 1 else f"--{name}"
        return f"{self.prefix}{name}"

    def namespaces(self,
                   index: Union[int, str] = 'len') -> Union[List[str], int]:
        """Get the namespaces at the given index or number of namespaces

        Args:
            index: The index or a length indicator

        Returns:
            The length of the namespaces or the namespaces at index.
        """
        if index == 'len':
            return len(self._namespaces)
        try:
            return self._namespaces[index]
        except IndexError:
            return []

    def full_names(self) -> List[str]:
        """make the names with full combinations of namespaces and terminals

        Since one can define a parameter like `n.arg` but namespace `n` can
        have aliases (i.e. `ns`). This makes sure the names of `n.arg` expands
        to `n.arg` and `ns.arg`

        Returns:
            The names with full combinations of namespaces and terminals
        """
        self.names = ['.'.join(prod)
                      for prod in product(*self._namespaces, self.terminals)]

    @property
    def is_positional(self) -> bool:
        """Tell if this parameter is positional

        Returns:
            True if it is, otherwise False.
        """
        return POSITIONAL in self.names

    def close(self) -> None:
        """Close up the parameter while scanning the command line

        We are mostly doing nothing, only if, say, param is bool and
        it was just hit, we should push a true value to it.
        """
        if self.hit:
            logger.warning("No value provided for argument %r", self.namestr())

    def overwrite_type(self, param_type: Optional[str]) -> Type['Param']:
        """Try to overwrite the type

        Only when param_type is not None and it's different from mine
        A new param will be returned if different

        Args:
            param_type: The type to overwrite

        Returns:
            Self when type not changed otherwise a new parameter with
                the given type
        """
        if param_type is None or param_type == self.typestr():
            return self

        if self.type_frozen:
            raise PyParamTypeError(f"Type of argument {self.namestr()!r} "
                                   "is not overwritable")
        logger.warning("Type changed from %r to %r for argument %r",
                       self.typestr(), param_type, self.namestr())
        return self.to(param_type)

    def consume(self, value: Any) -> bool:
        """Consume a value

        Args:
            Value: value to consume

        Returns:
            True if value was consumed, otherwise False
        """
        if self.hit or not self._stack:
            self.push(value)
            return True
        return False

    @property
    def desc(self) -> List[str]:
        """The formatted description using attributes and _kwargs"""
        format_data: dict = {
            key: val for key, val in self.__dict__.items()
            if (not key.startswith('_') and key != 'desc' and not callable(val))
        }
        format_data.update(self._kwargs)
        ret: List[str] = []
        for descr in self._desc:
            try:
                descr = descr.format(**format_data)
            except KeyError as kerr:
                raise PyParamException(
                    f'Description of {self.namestr()!r} is formatting '
                    'using kwargs from contructor. \n'
                    'If you have curly braces in it, which is not intended '
                    'for formatting, please escape it by replacing with '
                    '`{{` or `}}`:\n'
                    f'- desc: {descr}\n'
                    f'- key : {{... {str(kerr)[1:-1]} ...}}'
                ) from None
            else:
                ret.append(descr)
        return ret

    def name(self, which: str, with_prefix: bool = True) -> str:
        """Get the shortest/longest name of the parameter

        A name is ensured to be returned. It does not mean it is the real
        short/long name, but just the shortest/longest name among all the names

        Args:
            which: Whether get the shortest or longest name
                Could use `short` or `long` for short.
            with_prefix: Whether to include the prefix or not

        Returns:
            The shortest/longest name of the parameter
        """
        name: str = list(sorted(self.names,
                                key=len))[0 if 'short' in which else -1]
        return name if not with_prefix else self._prefix_name(name)

    def namestr(self, sep: str = ", ", with_prefix: bool = True) -> str:
        """Get all names connected with a separator.

        Args:
            sep: The separator to connect the names
            with_prefix: Whether to include the prefix or not
        Returns:
            the connected names
        """
        names: list = ['POSITIONAL' if name == POSITIONAL
                       else self._prefix_name(name) if with_prefix
                       else name for name in sorted(self.names, key=len)]
        return sep.join(names)

    def typestr(self) -> str:
        """Get the string representation of the type

        Returns:
            the string representation of the type
        """
        return self.type if not self.subtype else f"{self.type}:{self.subtype}"

    def usagestr(self) -> str:
        """Get the string representation of the parameter in the default usage
        constructor

        Returns:
            the string representation of the parameter in the default usage
        """
        # * makes sure it's not wrapped
        ret: str = self.name('long') + "*"
        ret += self.typestr().upper() if self.type_frozen else self.typestr()
        return ret

    def optstr(self) -> str:
        """Get the string representation of the parameter names and types
        in the optname section in help page

        Returns:
            the string representation of the parameter names and types
                in the optname section in help page
        """
        typestr: str = (self.typestr().upper()
                        if self.type_frozen else self.typestr())
        if not self.ns_param or not self.argname_shorten:
            # * makes sure it's not wrapped
            return (self.namestr() if self.is_help
                    else f"{self.namestr()}*<{typestr}>")

        ret: List[str] = []
        for term in sorted(self.terminals, key=len):
            ret.append(f"~<ns>.{term}")
        return ", ".join(ret) + f"*<{typestr}>"

    def to(self, to_type: str) -> Type['Param']:
        """Generate a different type of parameter using current settings

        Args:
            to_type: the type of parameter to generate

        Returns:
            the generated parameter with different type
        """
        # Type: Optional[str], Optional[str]
        main_type, sub_type = parse_type(to_type)
        klass: Callable = PARAM_MAPPINGS[main_type]
        param: Type['Param'] = klass(
            names=self.names,
            default=None,
            desc=self.desc,
            prefix=self.prefix,
            show=self.show,
            required=self.required,
            subtype=sub_type,
            callback=self.callback,
            complete_callback=self.complete_callback,
            argname_shorten=self.argname_shorten
        )
        param.ns_param = self.ns_param
        return param

    def copy(self) -> Type['Param']:
        """Copy a parameter so that it can be reused.

        Returns:
            The copy of the parameter
        """
        return self.__class__(
            names=self.names,
            default=self.default,
            desc=self.desc,
            prefix=self.prefix,
            show=self.show,
            required=self.required,
            subtype=self.subtype,
            callback=self.callback,
            type_frozen=self.type_frozen,
            complete_callback=self.complete_callback,
            argname_shorten=self.argname_shorten
        )

    @property
    def default_group(self) -> str:
        """Get the default group of the parameter

        Returns:
            the default group name
        """
        ret: str = "REQUIRED OPTIONS" if self.required else "OPTIONAL OPTIONS"
        return (ret if not self.ns_param
                else f"{ret} UNDER {self.ns_param.name('long')}")

    @property
    def desc_with_default(self) -> Optional[List[str]]:
        """If default is not specified in desc, just to add with the default
        value

        Returns:
            list of descriptions with default value added
        """
        if self.is_help:
            return self.desc

        if (self.required or (
                self.desc and
                any("Default:" in desc or "DEFAULT:" in desc
                    for desc in self.desc)
        )):
            return None if self.desc is None else self.desc[:]

        desc: List[str] = self.desc[:] if self.desc else ['']
        if desc[0] and not desc[0][-1:].isspace():
            desc[0] += " "

        default_str: str = str(self.default) or "''"
        desc[0] += f"Default: {default_str}"
        return desc

    def push(self, item: Any) -> None:
        """Push a value into the stack for calculating

        Returns:
            The item to be pushed
        """
        if self.hit is True and self._stack:
            logger.warning(
                "Previous value of argument %r is overwritten with %r.",
                self.namestr(), item
            )
            self._stack = []

        if self.hit or not self._stack:
            self._stack.append([])
        self._stack[-1].append(item)
        self.hit = False

    def __repr__(self) -> str:
        return (f'<{self.__class__.__name__}({self.namestr()} :: '
                f'{self.typestr()}) @ {id(self)}>')

    def _value(self) -> Any:
        """Get the organized value of this parameter

        Returns:
            The parsed value of thie parameter
        """
        if not self._stack:
            if self.required:
                raise PyParamValueError("Argument is required.")
            return self.default
        ret = self._stack[-1][0]
        self._stack = []
        return ret

    @property
    def value(self) -> Any:
        """Return the cached value if possible, otherwise calcuate one

        Returns:
            The cached value of this parameter or the newly calculated
                from stack
        """
        if self._value_cached is not None:
            return self._value_cached
        self._value_cached = self._value()
        return self._value_cached

    def apply_callback(self, all_values: Namespace) -> Any:
        """Apply the callback function to the value

        Args:
            all_values: The namespace of values of all parameters

        Returns:
            The value after the callback applied

        Raises:
            PyParamTypeError: When exceptions raised or returned from callback
        """
        if not callable(self.callback):
            return self.value
        try:
            try:
                val: Any = self.callback(self.value, all_values)
            except TypeError:
                val: Any = self.callback(self.value)
        except Exception as exc:
            raise PyParamTypeError(str(exc)) from None

        if isinstance(val, Exception):
            raise PyParamTypeError(str(val))
        return val

class ParamAuto(Param):
    """An auto parameter whose value is automatically casted"""
    type: str = 'auto'

    def _value(self) -> Any:
        """Cast value automatically"""
        return cast_to(super()._value(), 'auto')

class ParamInt(Param):
    """An int parameter whose value is automatically casted into an int"""
    type: str = 'int'
    type_aliases: List[str] = ['i']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = self.default or 0

    def _value(self) -> int:
        return int(super()._value())

class ParamFloat(Param):
    """A float parameter whose value is automatically casted into a float"""
    type: str = 'float'
    type_aliases: List[str] = ['f']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = self.default or 0.0

    def _value(self) -> float:
        return float(super()._value())

class ParamStr(Param):
    """A str parameter whose value is automatically casted into a str"""
    type: str = 'str'
    type_aliases: List[str] = ['s']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = self.default or ''

class ParamBool(Param):
    """A bool parameter whose value is automatically casted into a bool"""
    type: str = 'bool'
    type_aliases: List[str] = ['b', 'flag']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = self.default or False
        try:
            cast_to(str(self.default), 'bool')
        except (PyParamValueError, PyParamTypeError):
            raise PyParamValueError(
                "Default value of a count argument must be a bool value or "
                "a value that can be casted to a bool value."
            ) from None

    def usagestr(self) -> str:
        """Get the string representation of the parameter in the default usage
        constructor

        Returns:
            the string representation of the parameter in the default usage
        """
        return self.name('long')

    def optstr(self) -> str:
        """Get the string representation of the parameter names and types
        in the optname section in help page

        Returns:
            the string representation of the parameter names and types
                in the optname section in help page
        """
        if self.is_help:
            return self.namestr()
        typestr: str = (self.typestr().upper()
                        if self.type_frozen else self.typestr())

        if not self.ns_param or not self.argname_shorten:
            # * makes sure it's not wrapped'
            return f"{self.namestr()}*[{typestr}]"

        ret: List[str] = []
        for term in sorted(self.terminals, key=len):
            ret.append(f"~<ns>.{term}")
        return ", ".join(ret) + f"*[{typestr}]"

    def close(self) -> None:
        if self.hit is True:
            self.push('true')

    def consume(self, value: str) -> bool:
        """Should I consume given value?"""
        if not self.hit:
            return False

        try:
            cast_to(value, 'bool')
        except (PyParamValueError, PyParamTypeError):
            return False # cannot cast, don't consume
        else:
            self.push(value)
            return True

    def _value(self) -> bool:
        if not self._stack:
            return False

        ret = cast_to(self._stack[-1][0], 'bool')
        self._stack = []
        return ret

    def complete_value(
            self,
            current: str,
            prefix: str = ''
    ) -> Optional[Union[
        str, Iterator[Tuple[str]],
        Iterator[Tuple[str, str]], Iterator[Tuple[str, str, str]]
    ]]:
        """Get the completion candidates for the current parameter"""
        if self.is_help:
            return ''
        if callable(self.complete_callback):
            return super().complete_value(current, prefix)
        if current:
            trues: List[str] = ('True', 'true', 'TRUE', '1')
            falses: List[str] = ('False', 'false', 'FALSE', '0')
            ret: List[Tuple[str, str, str]] = []
            for cand in trues:
                if cand.startswith(current):
                    ret.append((f'{prefix}{cand}', 'plain', 'Value True'))

            for cand in falses:
                if cand.startswith(current):
                    ret.append((f'{prefix}{cand}', 'plain', 'Value False'))
            return ret
        return ''

class ParamCount(Param):
    """A bool parameter whose value is automatically casted into a bool"""
    type: str = 'count'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.default = self.default or 0
        if self.default != 0:
            raise PyParamValueError(
                "Default value of a count argument must be 0"
            )

        if len(self.name('short', with_prefix=False)) != 1:
            raise PyParamValueError(
                "Count argument must have a short name."
            )

        self._kwargs.setdefault('max', 0) # 0: no max
        if not isinstance(self._kwargs['max'], int) or self._kwargs['max'] < 0:
            raise PyParamValueError(
                "Argument 'max' for count argument must be a positive integer"
            )

    def close(self) -> None:
        if self.hit is True:
            self.push('1')

    def consume(self, value: str) -> bool:
        """Should I consume given parameter?"""
        if self.hit and value.isdigit():
            self.push(value)
            return True
        return False

    def _value(self) -> int:
        val = super()._value()

        retval = None
        if str(val).isdigit():
            retval = int(val)
        else:
            for name in self.names:
                if len(name) != 1:
                    continue
                if name * len(val) == val:
                    # -vvv => name: v, value: vv
                    # len(vv) = 2, but value should be 3
                    retval = len(val) + 1
                    break
        if retval is None:
            raise PyParamValueError("Expect repeated short names or "
                                    "an integer as count argument value.")
        if self._kwargs['max'] and retval > self._kwargs['max']:
            raise PyParamValueError(f"{retval} is greater than "
                                    f"the max of {self._kwargs['max']}.")
        return retval

    def complete_name(self, current: str) -> Iterator[Tuple[str, str]]:
        """Complete names for a count parameter

        Since we have -v, -vv, -vvv allowed for a count parameter, we need
        to put them in the completions, too.
        """
        # check if current is trying to do so
        for name in self.names:
            if len(name) != 1:
                continue
            # current is not like -vvv
            if self._prefix_name(name) + len(current[2:]) * name != current:
                continue
            # check the max
            value: int = len(current) - 1
            if self._kwargs['max'] and value >= self._kwargs['max']:
                # already max'ed, no further completion
                continue

            ncompletes: int = (min(self._kwargs['max'] - value, 2)
                               if self._kwargs['max'] else 2)
            for i in range(ncompletes):
                yield current + name * (i + 1), self.desc[0].splitlines()[0]
            break

        else:
            yield from super().complete_name(current)

class ParamPath(Param):
    """A path parameter whose value is automatically casted into a pathlib.Path
    """
    type: str = 'path'
    type_aliases: List[str] = ['p', 'file']

    def _value(self) -> Path:
        val: Optional[Path] = super()._value()
        return None if val is None else Path(val)

    def complete_value(
            self,
            current: str,
            prefix: str = ''
    ) -> Optional[Union[
        str, Iterator[Tuple[str]],
        Iterator[Tuple[str, str]], Iterator[Tuple[str, str, str]]
    ]]:
        """Generate file paths with given current prefix
        as completion candidates

        Args:
            current: The current word or prefix under cursor
        """
        if callable(self.complete_callback):
            return super().complete_value(current, prefix)
        return [(current, 'file', prefix)]

class ParamDir(ParamPath):
    """Subclass of ParamPath.

    It does not make any difference with pyparam. However, it works differently
    for completions. The completion items for this param will only give
    directories instead of all paths
    """
    type: str = 'dir'
    type_aliases: List[str] = []

    def complete_value(
            self,
            current: str,
            prefix: str = ''
    ) -> Optional[Union[
        str, Iterator[Tuple[str]],
        Iterator[Tuple[str, str]], Iterator[Tuple[str, str, str]]
    ]]:
        """Generate dir paths with given current prefix as completion candidates

        Args:
            current: The current word or prefix under cursor
        """
        if callable(self.complete_callback):
            return super().complete_value(current, prefix)
        return [(current, 'dir', prefix)]

class ParamPy(Param):
    """A parameter whose value will be ast.literal_eval'ed"""
    type: str = 'py'

    def _value(self) -> Any:
        return ast.literal_eval(str(super()._value()))

class ParamJson(Param):
    """A parameter whose value will be parsed as json"""
    type: str = 'json'
    type_aliases: List[str] = ['j']

    def _value(self) -> Any:
        val: Any = super()._value()
        return None if val is None else json.loads(str(val))

class ParamList(Param):
    """A parameter whose value is a list"""
    type: str = 'list'
    type_aliases: List[str] = ['l', 'a', 'array']
    complete_relapse: bool = True

    @classmethod
    def on_register(cls):
        """Also register reset type"""
        name = 'reset'
        aliases = ['r']
        all_names = [name] + aliases

        for nam in all_names:
            TYPE_NAMES[nam] = name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = self.default or []
        self._stack.append(self.default)

    def overwrite_type(self, param_type: Optional[str]) -> Type['Param']:
        """Deal with when param_type is reset"""
        if param_type == 'reset':
            self._stack = []
            return self
        return super().overwrite_type(param_type)

    def consume(self, value: str) -> bool:
        """Should I consume given parameter?"""
        self.push(value)
        return True

    def push(self, item: str):
        """Push a value into the stack for calculating"""
        if self.hit or not self._stack:
            self._stack.append([])
        self._stack[-1].append(item)
        self.hit = False

    def _value(self) -> List[Any]:
        """Get the value a list parameter"""
        ret = [cast_to(val, self.subtype) for sublist in self._stack
               for val in sublist]

        if self.required and not ret:
            raise PyParamValueError("Argument is required.")
        self._stack = []
        return ret

class ParamChoice(Param):
    """A bool parameter whose value is automatically casted into a bool"""
    type: str = 'choice'
    type_aliases: List[str] = ['c']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'choices' not in self._kwargs:
            raise PyParamValueError("Argument 'choices' is required "
                                    "for ParamChoice.")

        if not isinstance(self._kwargs['choices'], (list, tuple)):
            raise PyParamValueError("Argument 'choices' must be "
                                    "a list or a tuple.")

        if not self.required and self.default is None:
            self.default = self._kwargs['choices'][0]

    def _value(self) -> Any:
        val = super()._value()

        if val not in self._kwargs['choices']:
            raise PyParamValueError(
                f"{val} is not one of {self._kwargs['choices']}"
            )
        self._stack = []
        return val

    def complete_value(
            self,
            current: str,
            prefix: str = ''
    ) -> Optional[Union[
        str, Iterator[Tuple[str]],
        Iterator[Tuple[str, str]], Iterator[Tuple[str, str, str]]
    ]]:
        """Generate choices with given current prefix as completion candidates

        Args:
            current: The current word or prefix under cursor
        """
        if callable(self.complete_callback):
            return super().complete_value(current, prefix)

        ret: Iterator[Tuple[str]] = []
        for choice in self._kwargs['choices']:
            if choice.startswith(current):
                ret.append((f"{prefix}{choice}", ))
        return ret

class ParamNamespace(Param):
    """A pseudo parameter serving as a namespace for parameters under it

    So that it is possible to do:
    ```shell
    prog --namespace.arg1 1 --namespace.arg2 2
    ```
    """
    type: str = 'ns'
    type_aliases: List[str] = ['namespace']

    def __init__(self, *args, **kwargs):
        kwargs['desc'] = kwargs.get('desc') or [
            "Works as a namespace for other arguments.",
            "Never pass a value directly to it."
        ]
        kwargs['default'] = None
        super().__init__(*args, **kwargs)
        self._stack = OrderedDiot() # for my decendents

    @property
    def default_group(self) -> str:
        """Get the default group of the parameter"""
        ret: str = ('REQUIRED'
                    if any(param.required for param in self._stack.values())
                    else 'OPTIONAL')
        return (f'{ret} OPTIONS' if not self.ns_param
                else f'{ret} OPTIONS UNDER {self.ns_param.name("long")}')

    @property
    def desc_with_default(self) -> Optional[List[str]]:
        """Namespace parameters do not have a default value"""
        return self.desc[:]

    def decendents(self, show_only=False) -> List[Type['Param']]:
        """Get all decendents of this namespace parameter

        Args:
            show_only: Load params with show=True only?

        Returns:
            The decendents of this namespace parameter
        """
        ret: List[Type['Param']] = []
        ret_append: Callable = ret.append
        for param in self._stack.values():
            # don't skip entire ns parameter
            if isinstance(param, ParamNamespace):
                if (not show_only or param.show) and param not in ret:
                    ret.append(param)
                ret.extend(param.decendents(show_only))
                continue
            if show_only and not param.show:
                continue
            if param not in ret:
                ret_append(param)
        return ret

    def consume(self, value: str) -> bool:
        """Should I consume given parameter?"""
        return False

    def get_param(self, name: str, depth: int = 0) -> Optional[Type["Param"]]:
        """Get the paraemeter by given name

        This parameter is like '-a', and the name can be 'a.b.c.d'

        Args:
            name: The name of the parameter to get
            depth: The depth

        Returns:
            The parameter we get with the given name
        """
        parts: List[str] = name.split('.')
        if depth < len(parts) - 1:
            part = parts[depth + 1]
            if part not in self._stack:
                return None
            if not isinstance(self._stack[part], ParamNamespace):
                return self._stack[part]
            return self._stack[part].get_param(name, depth + 1)
        return None

    # pylint: disable=arguments-differ
    def push(self, item: Type["Param"], depth: int = 0) -> None:
        """Push the parameter under this namespace.

        We are not pushing any values to this namespace, but pushing
        parameters that are under it.
        """
        # check if item's namespaces at depth contain only names of this
        # ns parameter
        if set(item.namespaces(depth)) - set(self.terminals):
            raise PyParamValueError(
                "Parameter names should only contain namespace names "
                "that belong to the same namespace parameter."
            )

        # set the names with all possible combination
        item._namespaces[depth] = self.terminals #[:]
        item.full_names()

        # check if we have nested namespaces
        if 0 <= depth < item.namespaces('len') - 1:
            # see if sub-ns exists
            subns_name = item.namespaces(depth + 1)[0]
            if subns_name in self._stack:
                subns = self._stack[subns_name]
            else:
                subns = ParamNamespace([
                    '.'.join(prod) for prod in product(
                        *(item.namespaces(i) for i in range(depth+2))
                    )
                ])
                subns.ns_param = self

                for name in subns.terminals:
                    self._stack[name] = subns
            subns.push(item, depth + 1)
        else:
            item.ns_param = self
            for term in item.terminals:
                self._stack[term] = item
    # pylint: enable=arguments-differ

    def _value(self) -> Namespace:
        val = Namespace()
        for param_name, param in self._stack.items():
            if param_name not in val:
                for name in param.terminals:
                    val[name] = param.value
        return val

    def copy(self) -> Type['Param']:
        """Copy a parameter so that it can be reused.

        Returns:
            The copy of the parameter
        """
        copied = super().copy()
        copied._stack = OrderedDiot([(key, param.copy())
                                     for key, param in self._stack.items()])
        return copied

    def apply_callback(self, all_values: Namespace) -> Any:
        ns_callback_applied = Namespace()
        for param_name, param in self._stack.items():
            if param_name not in ns_callback_applied:
                for name in param.terminals:
                    ns_callback_applied[name] = param.apply_callback(all_values)

        if not callable(self.callback):
            return ns_callback_applied

        try:
            try:
                val = self.callback(ns_callback_applied, all_values)
            except TypeError:
                val = self.callback(ns_callback_applied)
        except Exception as exc:
            raise PyParamTypeError(str(exc)) from None

        if isinstance(val, Exception):
            raise PyParamTypeError(str(val))
        return val

def register_param(param: Param) -> None:
    """Register a parameter class

    Args:
        param: The param to register
            A param class should include a type
            You can also define type alias for a param type
    """
    for alias in param.type_aliases + [param.type]:
        if alias in TYPE_NAMES:
            raise PyParamNameError('Type name has already been '
                                   f'registered: {alias}')
        TYPE_NAMES[alias] = param.type

    PARAM_MAPPINGS[param.type] = param
    param.on_register()

for param_class in (ParamAuto, ParamInt, ParamStr, ParamFloat, ParamBool,
                    ParamCount, ParamPath, ParamDir, ParamPy, ParamJson,
                    ParamList, ParamChoice, ParamNamespace):
    register_param(param_class)



"""Definition of Params"""
# pylint: disable=too-many-lines
import sys
from typing import Optional, Type, Union, List, Callable, Any, Tuple, Dict
from pathlib import Path
from diot import OrderedDiot, Diot
import rich
from simpleconf import Config, LOADERS
from .utils import (
    always_list,
    Namespace,
    logger,
    parse_type,
    type_from_value,
    parse_potential_argument
)
from .defaults import (POSITIONAL,
                       PARAMS as PARAMS_DEFAULT,
                       PARAM as PARAM_DEFAULT)
from .param import PARAM_MAPPINGS
from .help import HelpAssembler, ProgHighlighter
from .completer import Completer
from .exceptions import PyParamNameError

class Params(Completer): # pylint: disable=too-many-instance-attributes
    """Params, served as root params or subcommands

    Args:
        names: The names of this command if served as a command
        desc: The description of the command.
            This will be finally compiled into a list if a string is given.
            The difference is, when displayed on help page, the string will
            be wrapped by textwrap automatically. However, each element in
            a given list will not be wrapped.
        prog: The program name
        help_keys: The names to bring up the help information
        help_cmds: The help command names to show help of other
            subcommands
        help_on_void: Whether to show help when no arguments provided
        help_callback: A function to modify the help page
        help_modifier: A callback function to modify the help param/command
        prefix: The prefix for the arguments
            (see attribute `Params.prefix`)
        arbitrary: Whether to parse the command line arbitrarily
        theme (str|rich.theme.Theme): The theme to render the help page
        usage: Some example usages

    Attributes:
        desc: The description of the command.
        prog: The program name. Default: `sys.argv[0]`
        help_keys: The names to bring up the help information.
        help_cmds: The names of help subcommands to bring up
            help information of subcommands
        help_on_void: Whether show help when there is not arguments
            provided
        usage: The usages of this program
        prefix: The prefix for the arguments on command line
            - `auto`: Automatically determine the prefix for each argument.
                Basically, `-` for short options, and `--` for long.
                Note that `-` for `-vvv` if `v` is a count option
        arbitrary: Whether parsing the command line arbitrarily
        theme (rich.theme.Theme|str): The theme for the help page
        names: The names of the commands if this serves as sub-command
        params: The ordered dict of registered parameters
        commands: The ordered dict of registered commands
        param_groups: The ordered dict of parameter groups
        command_groups: The ordered dict of command groups
        asssembler: The asssembler to assemble the help page
    """
    def __init__(self, # pylint: disable=too-many-arguments
                 names: Optional[Union[str, List[str]]] = None,
                 desc: Optional[Union[List[str], str]] = None,
                 prog: Optional[str] = None,
                 help_keys: Union[str, List[str]] = None,
                 help_cmds: Union[str, List[str]] = None,
                 help_on_void: bool = None,
                 help_callback: Optional[Callable] = None,
                 help_modifier: Optional[Callable] = None,
                 prefix: str = 'auto',
                 arbitrary: bool = False,
                 theme: str = 'default',
                 usage: Optional[Union[str, List[str]]] = None) -> None:
        self.names: List[str] = always_list(names) if names else []
        self.desc: List[str] = (PARAMS_DEFAULT.desc if desc is None
                                else always_list(desc,
                                                 strip=False,
                                                 split=False))
        self._prog: str = sys.argv[0] if prog is None else prog
        self.help_keys: List[str] = (PARAMS_DEFAULT.help_keys
                                     if help_keys is None
                                     else always_list(help_keys))
        self.help_cmds: List[str] = (PARAMS_DEFAULT.help_cmds
                                     if help_cmds is None
                                     else always_list(help_cmds))
        self.help_on_void: bool = (PARAMS_DEFAULT.help_on_void
                                   if help_on_void is None
                                   else help_on_void)
        self.usage: Optional[List[str]] = (
            None if usage is None
            else always_list(usage, strip=True, split='\n')
        )
        self.prefix: str = prefix
        self.arbitrary: bool = arbitrary
        self.theme: Union[str, rich.theme.Theme] = theme

        self.params: OrderedDiot = OrderedDiot()
        self.commands: OrderedDiot = OrderedDiot()

        self.param_groups: OrderedDiot = OrderedDiot()
        self.command_groups: OrderedDiot = OrderedDiot()

        self.help_modifier = help_modifier

        self.assembler: HelpAssembler = HelpAssembler(
            self.prog, theme, help_callback
        )
        super().__init__()

    @property
    def prog(self) -> str:
        """Get the program name"""
        return self._prog

    @prog.setter
    def prog(self, value: str):
        """Set the program name and update the help assembler

        Args:
            value: The new program name
        """
        self._prog = value
        self.assembler.console.meta.prog = value
        self.assembler.console.meta.highlighters.prog = ProgHighlighter(
            value
        )

    def name(self, which: str = 'short') -> str:
        """Get the shortest/longest name of the parameter

        A name is ensured to be returned. It does not mean it is the real
        short/long name, but just the shortest/longest name among all the names

        Args:
            which: Whether get the shortest or longest name
                Could use `short` or `long` for short.

        Returns:
            The shortest/longest name of the parameter
        """
        return list(sorted(self.names, key=len))[0 if 'short' in which else -1]

    def namestr(self, sep: str = ", ") -> str:
        """Get all names connected with a separator.

        Args:
            sep: The separator to connect the names

        Returns:
            the connected names
        """
        return sep.join(sorted(self.names, key=len))

    def get_param(self, name: Optional[str]) -> Optional[Type["Param"]]:
        """Get the parameter by name

        If the parameter is under a namespace, try to get it via the namespace

        Args:
            name: The name of the parameter to get (without prefix)

        Returns:
            The parameter, None if failed
        """
        if name is None:
            return None
        if '.' not in name:
            return self.params.get(name)

        ns_param: Optional["ParamNamespace"] = self.params.get(
            name.split('.', 1)[0]
        )
        if not ns_param:
            return None
        ret = ns_param.get_param(name)
        return None if name not in ret.names else ret

    def get_command(self, name: str) -> Optional["Params"]:
        """Get the command object

        Args:
            name: The name of the command to get

        Returns:
            The command object, None if failed.
        """
        return self.commands.get(name)

    def _set_param(self, param: Type["Param"]) -> None:
        """Set the parameter

        When a paraemeter's type is overwritten, we need to replace it with
        the new one in the pool (either self.params or the namespace parameter
        that holds it).
        """
        if param.namespaces(0):
            param.ns_param.push(param, -1)
        else:
            for name in param.names:
                self.params[name] = param

    def add_param(self, # pylint: disable=too-many-arguments
                  names: Union[str, List[str], Type["Param"]],
                  default: Any = None,
                  # pylint: disable=redefined-builtin
                  type: Optional[str] = None,
                  desc: Union[str, List[str]] = None,
                  show: bool = None,
                  required: bool = None,
                  callback: Optional[Callable] = None,
                  group: Optional[str] = None,
                  force: bool = False,
                  type_frozen: bool = None,
                  argname_shorten: bool = None,
                  complete_callback: Optional[Callable] = None,
                  **kwargs) -> Type["Param"]:
        """Add an argument

        Args:
            names: names of the argument or a parameter defined somewhere else
                For example, in case we want to reuse a parameter
                ```python
                param = cmd1.add_param('n,name')
                # reuse it:
                cmd2.add_param(param)
                # other arguments will be ignored, except force
                ```

            default: The default value for the argument
            type: The type of the argument
                Including single value type and complex one
                - Single value types:
                    auto, int, str, float, bool, count, py, json, reset
                - Complex value types:
                    list[<single value type>], ns
            desc: The description of the argument
                This will be finally compiled into a list if a string is given.
                The difference is, when displayed on help page, the string will
                be wrapped by textwrap automatically. However, each element in
                a given list will not be wrapped.
            show: Whether this should be shown on help page.
            required: Whether this argument is required from
                the command line
            callback: Callback to convert parsed values
            group: The group this parameter belongs to.
                Arguments will be grouped by this on the help page.
            force: Whether to force adding parameter if it exists
            type_frozen: Whether allow type overwritting from
                the commone line
            argname_shorten: Whether show shortened name for parameter
                under namespace parameter
            complete_callback: The callback for complete the values of the
                parameter
            **kwargs: Additional keyword arguments

        Raises:
            PyParamNameError: When parameter exists and force is false

        Return:
            Param: The added parameter
        """
        # pylint: disable=too-many-locals
        if isinstance(names, (str, list)):
            names: List[str] = always_list(names)
            if type is None:
                if POSITIONAL in names and default is None:
                    type = 'list'
                else:
                    type = type_from_value(default)

            # Type: Optional[str], Optional[str]
            maintype, subtype = parse_type(type.__name__
                                           if callable(type)
                                           else PARAM_DEFAULT.type
                                           if type is None
                                           else type)

            param: Type['Param'] = PARAM_MAPPINGS[maintype](
                names=names,
                default=PARAM_DEFAULT.default if default is None else default,
                desc=PARAM_DEFAULT.desc if desc is None else always_list(
                    desc, strip=False, split=False
                ),
                prefix=self.prefix,
                show=PARAM_DEFAULT.show if show is None else show,
                required=(PARAM_DEFAULT.required if required is None
                          else required),
                subtype=subtype,
                type_frozen=(PARAM_DEFAULT.type_frozen if type_frozen is None
                             else type_frozen),
                callback=callback,
                argname_shorten=(PARAM_DEFAULT.argname_shorten
                                 if argname_shorten is None
                                 else argname_shorten),
                complete_callback=complete_callback,
                **kwargs
            )
        else:
            param = names.copy()

        if any(name in self.help_keys for name in param.names):
            param.is_help = True

        if param.namespaces(0):
            # add namespace parameter automatically
            if param.namespaces(0)[0] not in self.params:
                self.add_param(param.namespaces(0), type="ns")

            ns_param: "ParamNamespace" = self.params[param.namespaces(0)[0]]
            ns_param.push(param)

        else:
            for name in param.names:
                # check if parameter has been added
                if not force and (name in self.params or name in self.commands):
                    raise PyParamNameError(
                        f"Argument {name!r} has already been added."
                    )
                self.params[name] = param

        group = group or param.default_group
        # leave the parameters here under namespace to have flexibility
        # assigning different groups
        groups: List[Type['Param']] = self.param_groups.setdefault(group, [])

        # any parameter with param.names hasn't been added
        if not any(set(param.names) & set(prm.names) for prm in groups):
            groups.append(param)
        return param

    def add_command(self, # pylint: disable=too-many-arguments
                    names: Union["Params", str, List[str]],
                    desc: Optional[Union[str, List[str]]] = 'No description',
                    help_keys: Union[str, List[str]] = '__inherit__',
                    help_cmds: Union[str, List[str]] = '__inherit__',
                    help_on_void: Union[str, bool] = '__inherit__',
                    help_callback: Optional[Callable] = None,
                    help_modifier: Optional[Callable] = None,
                    prefix: str = '__inherit__',
                    arbitrary: Union[str, bool] = '__inherit__',
                    theme: Union[str, rich.theme.Theme] = '__inherit__',
                    usage: Optional[Union[str, List[str]]] = None,
                    group: Optional[str] = None,
                    force: bool = False) -> "Params":
        """Add a sub-command

        Args:
            names: list of names of this command
                It can also be a Params object that served as a subcommand
            desc: description of this command
            help_keys: help key for bring up help for this command
            help_cmds: help command for printing help for other
                sub-commands of this command
            help_on_void: whether printing help when no arguments passed
            help_callback: callback to manipulate help page
            prefix: prefix for arguments for this command
            arbitray: whether do arbitray Parsing
            theme: The theme of help page for this command
            usage: Usage for this command
            group: Group of this command
            force: Force adding when command exists already.

        Returns:
            The added command
        """
        # pylint: disable=too-many-locals
        command: Optional["Params"] = None
        if isinstance(names, Params):
            command = names
            command.prog = (
                f"{self.prog}{' [OPTIONS]' if self.params else ''} "
                f"{command.name('long')}"
            )
        else:
            names: List[str] = always_list(names)
            command: "Params" = Params(
                desc=desc,
                prog=(f"{self.prog}{' [OPTIONS]' if self.params else ''} "
                      f"{sorted(names, key=len)[-1]}"),
                help_keys=(self.help_keys if help_keys == '__inherit__'
                           else help_keys),
                help_cmds=(self.help_cmds if help_cmds == '__inherit__'
                           else help_cmds),
                help_on_void=(self.help_on_void if help_on_void == '__inherit__'
                              else help_on_void),
                help_callback=help_callback,
                help_modifier=help_modifier,
                prefix=(self.prefix if prefix == '__inherit__'
                        else prefix),
                arbitrary=(self.arbitrary
                           if arbitrary == '__inherit__'
                           else arbitrary),
                theme=(self.theme if theme == '__inherit__'
                       else theme),
                usage=usage,
                names=names
            )
        for cmd in command.names:
            # check if command has been added
            if not force and (cmd in self.params or cmd in self.commands):
                raise PyParamNameError(
                    f"Command {cmd!r} has already been added."
                )
            self.commands[cmd] = command

        group = group or "COMMANDS"
        groups: List['Params'] = self.command_groups.setdefault(group, [])

        if not any(set(command.names) & set(cmd.names) for cmd in groups):
            groups.append(command)
        return command

    def print_help(self, exit_code: int = 1) -> None:
        """Print the help information and exit

        Args:
            exit_code: The exit code or False to not exit
        """
        self.assembler.assemble(self)
        self.assembler.printout()
        if exit_code is not False:
            sys.exit(exit_code)

    def values(self,
               namespace: Optional[Namespace] = None) -> Optional[Namespace]:
        """Get a namespace of all parameter name => value pairs or attach them
        to the given namespace

        Args:
            namespace: The namespace for the values to attach to.

        Returns:
            the namespace with values of all parameter
                name-value pairs
        """
        ns_no_callback: Namespace = Namespace()
        for param_name, param in self.params.items():
            if param.is_help or param_name in ns_no_callback:
                continue
            try:
                value: Any = param.value
            except (TypeError, ValueError) as pve:
                logger.error("%r: %s", param.namestr(), pve)
                self.print_help()
            else:
                for name in param.names:
                    ns_no_callback[name] = value

        if namespace is None:
            namespace = Namespace()

        for param_name, param in self.params.items():
            if param.is_help or param_name in namespace:
                continue
            try:
                value: Any = param.apply_callback(ns_no_callback)
            except (TypeError, ValueError) as pte:
                logger.error("%r: %s", param.namestr(), pte)
                self.print_help()
            else:
                for name in param.names:
                    setattr(namespace, name, value)
        return namespace

    def parse(self,
              args: Optional[List[str]] = None) -> Namespace:
        """Parse the arguments from the command line

        Args:
            args: The arguments to parse

        Return:
            Namespace: The namespace of parsed arguments
        """
        # add help options here so that user can disable or
        # change it before parsing
        self.add_param(self.help_keys,
                       type='bool',
                       desc='Print help information for this command',
                       force=True)

        help_cmd: Optional["Params"] = None
        if self.commands:
            help_cmd = self.add_command(
                self.help_cmds,
                desc='Print help of sub-commands',
                force=True
            )
            help_cmd.add_param(
                POSITIONAL,
                type='str',
                default="",
                desc="Command name to print help for"
            )

        if callable(self.help_modifier):
            self.help_modifier(self.get_param(self.help_keys[0]), help_cmd)

        if args is None:
            # enable completion only when we are trying to parse sys.argv
            if self.comp_shell:
                print("\n".join(self.complete()))
                sys.exit(0)
            args = sys.argv[1:]

        if not args and self.help_on_void:
            self.print_help()

        namespace: Namespace = Namespace()
        self._parse(args, namespace)

        if self.commands and not namespace.__command__:
            logger.error('No command given.')
            self.print_help()
        # run help subcommand
        elif (
                namespace.__command__ in self.help_cmds and
                len(self.commands) > 1 # together with help command
        ):
            command_passed = namespace[namespace.__command__][POSITIONAL]
            if not command_passed:
                self.print_help()
            elif command_passed not in self.commands:
                logger.error('Unknown command: %r', command_passed)
                self.print_help()
            else:
                self.commands[command_passed].print_help()
        return namespace

    def _parse(self, # pylint: disable=too-many-branches
               args: List[str],
               namespace: Namespace) -> None:
        """Parse the arguments from the command line

        Args:
            args: The arguments to parse
            namespace: The namespace for parsed arguments to
                attach to.
        """
        logger.debug("Parsing %r", args)

        if not args: # help_on_void = False
            self.values(namespace)
            return

        prev_param: Optional[Type['Param']] = None
        for i, arg in enumerate(args):
            logger.debug("- Parsing item %r", arg)
            # Match the arg with defined parameters
            # If arbitrary, non-existing parameters will be created on the fly
            # This means
            # 1. if param_name is None
            #    arg is not a parameter-like format (ie. -a, --arg)
            #    then param_value == arg
            # 2. if param_name is not None, arg is parameter-like
            #    With arbitrary = True, parameter will be created on the fly
            # 3. if arg is like --arg=1, then param_value 1 is pushed to param.
            # Type: Optional[Type['Param']], Optional[str], Optional[str], Any
            param, param_name, param_type, param_value = self._match_param(arg)
            logger.debug("  Previous: %r", prev_param)
            logger.debug("  Matched: %r, name=%s, type=%s, value=%r",
                         param, param_name, param_type, param_value)
            # as long as the help argument hit
            if param_name in self.help_keys or (param and param.is_help):
                self.print_help()

            if param:
                if prev_param:
                    logger.debug("  Closing previous argument")
                    prev_param.close()
                prev_param = param

            elif prev_param: # No param
                if param_name is not None:
                    logger.warning("Unknown argument: %r, skipped", arg)
                elif not prev_param.consume(param_value):
                    # If value cannot be consumed, let's see if it
                    # 1. hits a command
                    # 2. hits the start of positional arguments
                    # Type: Optional[Type['Param']], Optional[str]
                    prev_param.close()
                    prev_param, matched = self._match_command_or_positional(
                        prev_param, param_value, args[(i+1):], namespace
                    )
                    if matched == 'command':
                        break
                    if matched == 'positional':
                        continue
                    if param_value is not None:
                        logger.warning(
                            "Unknown value: %r, skipped", param_value
                        )
                else:
                    logger.debug("  Param %r consumes %r",
                                 prev_param.namestr(), param_value)
            else: # neither
                # Type: Optional[Type['Param']], Optional[str]
                prev_param, matched = self._match_command_or_positional(
                    prev_param, param_value, args[(i+1):], namespace
                )
                if matched == 'command':
                    break
                if matched == 'positional':
                    continue
                if param_value is not None:
                    logger.warning("Unknown value: %r, skipped", param_value)

        if prev_param:
            logger.debug("  Closing final argument: %r", prev_param.namestr())
            prev_param.close()

        self.values(namespace)

    def _match_command_or_positional(
            self,
            prev_param: Optional[Type['Param']],
            arg: str,
            rest_args: List[str],
            namespace: Namespace
    ) -> Tuple[Optional[Type['Param']], Optional[str]]:
        """Check if arg hits a command or a positional argument start

        Args:
            prev_param: The previous parameter
            arg: The current argument item
            rest_args: The remaining argument items

        Returns:
            tuple (Param, str):
                - A parameter if we create a new one here
                    (ie, a positional parameter)
                - 'command' when arg hits a command or 'positional' when it hits
                    the start of a positional argument. Otherwise, None.
        """
        if prev_param and prev_param.is_positional:
            logger.debug("  Hit positional argument")
            prev_param.push(arg)
            return prev_param, 'positional'

        if arg not in self.commands:
            # any of the rest args matches is argument-like then
            # this should not hit the start of positional argument
            for rest_arg in rest_args:
                if self.prefix != 'auto' and rest_arg.startswith(self.prefix):
                    break

                if self.prefix == 'auto' and rest_arg[:1] == '-':
                    if len(rest_arg) <= 2 or (
                            rest_arg[:2] == '--' and len(rest_arg) > 3
                    ):
                        break
            else:
                logger.debug("  Hit start of positional argument")
                if self.arbitrary and POSITIONAL not in self.params:
                    self.add_param(POSITIONAL)
                if POSITIONAL in self.params:
                    self.params[POSITIONAL].hit = True
                    self.params[POSITIONAL].push(arg)
                    return self.params[POSITIONAL], 'positional'

        if prev_param:
            prev_param.close()

        if self.arbitrary and arg not in self.commands:
            self.add_command(arg)

        if arg not in self.commands:
            return None, None

        logger.debug("* Hit command: %r", arg)
        command: "Params" = self.commands[arg]
        namespace.__command__ = arg
        parsed: Namespace = command.parse(rest_args)
        for name in command.names:
            namespace[name] = parsed

        return None, 'command'

    def _match_param(self, arg: str) -> Tuple[
            Optional[Type['Param']],
            Optional[str], Optional[str], Optional[str]
    ]:
        """Check if arg matches any predefined parameters. With
        arbitrary = True, parameters will be defined on the fly.

        And then do all the preparation for the matched parameter, including
        overwrite the type and push the attached value, such as '--arg=1'

        Args:
            arg: arg to check

        Returns:
            The matched parameter, parameter name,
                type and unpushed value if matched.
                Otherwise, None, param_name, param_type and arg itself.
        """
        # Type: Optional[str], Optional[str], Optional[str]
        param_name, param_type, param_value = parse_potential_argument(
            arg, self.prefix
        )
        # parse -arg as -a rg only applicable with prefix auto and -
        # When we didn't match any argument-like
        # with allow_attached=False
        # Or we matched but it is not defined
        name_with_attached: Optional[str] = None
        if not param_type and self.prefix == 'auto':
            # then -a1 will be put in param_value, as if `a1` is a name,
            # it should be --a1
            name_with_attached = (
                param_value
                if (param_name is None and
                    param_value and param_value[:1] == '-' and
                    param_value[1:2] != '-')
                else None
            )

        elif not param_type and len(self.prefix) <= 1:
            # say prefix = '+'
            # then `a1` for `+a1` will be put as param_name, since
            # there is no restriction on name length
            name_with_attached = (
                self.prefix + param_name
                if param_name and param_name[:1] != self.prefix
                else None
            )

        # we cannot find a parameter with param_name
        # check if there is any value attached
        if name_with_attached and not self.get_param(param_name):
            # Type: Optional[str], Optional[str], Optional[str]
            param_name2, param_type2, param_value2 = (
                parse_potential_argument(
                    name_with_attached, self.prefix, allow_attached=True
                )
            )
            # Use them only if we found a param_name2 and
            # arbitrary: not previous param_name found
            # otherwise: parameter with param_name2 exists
            if param_name2 is not None and (
                    (self.arbitrary and param_name is None) or
                    self.get_param(param_name2)
            ):
                param_name, param_type, param_value = (
                    param_name2, param_type2, param_value2
                )

        # create the parameter for arbitrary
        if (self.arbitrary and
                param_name is not None and
                not self.get_param(param_name)):
            self.add_param(param_name, type=param_type)

        param: Optional[Type['Param']] = self.get_param(param_name)
        if not param:
            return None, param_name, param_type, param_value

        param_maybe_overwritten: Type['Param'] = param.overwrite_type(
            param_type
        )
        if param_maybe_overwritten is not param:
            self._set_param(param_maybe_overwritten)
            param = param_maybe_overwritten

        param.hit = True
        if param_value is not None:
            param.push(param_value)
        return param, param_name, param_type, param_value

    def _all_params(self, show_only=False) -> List[Type['Param']]:
        """All parameters under this command

        self.params don't have all of them, since namespaced ones are under
        namespace paramters
        """
        ret: List[Type['Param']] = []
        ret_append: Callable = ret.append
        for param in self.params.values():
            if param.type == 'ns':
                if (not show_only or param.show) and param not in ret:
                    ret.append(param)
                ret.extend(param.decendents(show_only))

            elif show_only and not param.show:
                continue
            elif param not in ret:
                ret_append(param)
        return ret

    def _to_dict_params(self) -> Diot:
        """Load all parameters into a dictionary"""
        ret = Diot()
        param_groups = {}
        for group, param_list in self.param_groups.items():
            for param in param_list:
                param_groups[param.names[0]] = group

        for param in self._all_params():
            param_name: str = ("POSITIONAL" if param.names[0] == POSITIONAL
                               else param.names[0])
            ret[param_name] = Diot()
            param_dict: Diot = ret[param_name]
            if len(param.names) > 1:
                param_dict.aliases = ["POSITIONAL" if name == POSITIONAL
                                      else name for name in param.names[1:]]
            if param.default != PARAM_DEFAULT.default:
                param_dict.default = param.default
            if param.typestr() != PARAM_DEFAULT.type:
                param_dict.type = param.typestr()
            if param.desc != PARAM_DEFAULT.desc:
                param_dict.desc = param.desc
            if param.show != PARAM_DEFAULT.show:
                param_dict.show = param.show
            if param.required != PARAM_DEFAULT.required:
                param_dict.required = param.required
            if param.type_frozen != PARAM_DEFAULT.type_frozen:
                param_dict.type_frozen = param.type_frozen
            if param.argname_shorten != PARAM_DEFAULT.argname_shorten:
                param_dict.argname_shorten = param.argname_shorten
            param_dict.group = group = param_groups[param.names[0]]
            param_dict |= param._kwargs

        return ret

    def _to_dict_commands(self) -> Diot:
        """Load all commands into a dictionary"""
        ret = Diot()
        command_groups = {}
        for group, command_list in self.command_groups.items():
            for command in command_list:
                for name in command.names:
                    command_groups[name] = group

        for command_name, command in self.commands.items():
            if any(name in ret for name in command.names):
                continue
            ret[command_name] = Diot()
            cmd_dict: Diot = ret[command_name]
            if len(command.names) > 1:
                cmd_dict.aliases = [name for name in command.names
                                    if name != command_name]
            if command.desc != PARAMS_DEFAULT.desc:
                cmd_dict.desc = command.desc
            if command.help_keys != PARAMS_DEFAULT.help_keys:
                cmd_dict.help_keys = command.help_keys
            if command.help_cmds != PARAMS_DEFAULT.help_cmds:
                cmd_dict.help_cmds = command.help_cmds
            if command.help_on_void != PARAMS_DEFAULT.help_on_void:
                cmd_dict.help_on_void = command.help_on_void
            if command.prefix != PARAMS_DEFAULT.prefix:
                cmd_dict.prefix = command.prefix
            if command.arbitrary != PARAMS_DEFAULT.arbitrary:
                cmd_dict.arbitrary = command.arbitrary
            if command.theme != PARAMS_DEFAULT.theme:
                cmd_dict.theme = command.theme
            if command.usage != PARAMS_DEFAULT.usage:
                cmd_dict.usage = command.usage
            cmd_dict.group = command_groups[command_name]
            cmd_dict |= command.to_dict()

        return ret

    def to_dict(self) -> Diot:
        """Save the parameters/commands to file.

        This is helpful if the parameters/commands take time to load. Once can
        cache this to a file, and load it from it using `from_file`.

        Returns:
            The complied Diot of parameters and commands
        """
        # load all parameters and commands
        return Diot(params=self._to_dict_params(),
                    commands=self._to_dict_commands())

    def to_file(self,
                path: Union[str, Path],
                cfgtype: Optional[str] = None) -> None:
        """Save the parameters/commands to file.

        This is helpful if the parameters/commands take time to load. Once can
        cache this to a file, and load it from it using `from_file`.

        Args:
            path: The path to the file
            cfgtype: The type of the file
                If not given, will inferred from the suffix of the path
                Supports one of yaml, toml and json
        """
        loaded: Diot = self.to_dict()
        if not isinstance(path, Path):
            path = Path(path)
        if not cfgtype:
            if path.suffix in ('.yml', '.yaml'):
                cfgtype = 'yaml'
            elif path.suffix == '.toml':
                cfgtype = 'toml'
            elif path.suffix == '.json':
                cfgtype = 'json'
            else:
                cfgtype = path.suffix
        if cfgtype not in ('yaml', 'json', 'toml'):
            raise ValueError('File type not supported: %s' % cfgtype)

        if cfgtype == 'yaml':
            loaded.to_yaml(path)
        elif cfgtype == 'json':
            loaded.to_json(path)
        else:
            loaded.to_toml(path)

    def from_file(self,
                  filename: Union[str, dict],
                  filetype: Optional[str] = None,
                  show: bool = True,
                  force: bool = False) -> None:
        """Load parameters from file

        We support 2 types for format to load the parameters.

        - express way, which has some limitations:
            1. no command definition;
            2. no namespace parameters;
        ```toml
        arg = 1 # default value
        "arg.desc" = "An argument" # description
        # other attributes
        ```
        or
        ```toml
        [arg]
        default = 1
        desc = "An argument"
        ```

        - full specification
        ```toml
        [params.arg]
        default = 1
        desc = "An argument"
        [commands.command]
        desc = "A subcommand"

          [commands.command.params.arg]
          default = 2
          desc = "An argument for command"
        ```

        Args:
            filename: path to the file
            filetype: The type of the file. If None, will infer from the
                filename. Supported types: ini, cfg, conf, config, yml, yaml,
                json, env, osenv, toml
            show: The default show value for parameters in the file
            force: Whether to force adding params/commands
        """
        config: Config = Config(with_profile=False)
        config._load(filename, factory=LOADERS.get(filetype))
        self.from_dict(config, show=show, force=force)

    def _from_dict_with_sections(self,
                                 dict_obj: Dict[str, Dict[str, dict]],
                                 show: bool = True,
                                 force: bool = False) -> None:
        """Load from the dict with 'params' and/or 'commands' sections"""

        param_section: Dict[str, dict] = dict_obj.get('params', {})
        # Type: str, dict
        for param_name, param_attrs in param_section.items():
            param_attrs.setdefault('show', show)
            names: List[str] = always_list(param_attrs.pop('aliases', []))
            names.insert(0, param_name)
            self.add_param([POSITIONAL if name == 'POSITIONAL'
                            else name for name in names],
                           **param_attrs, force=force)

        command_section: Dict[str, dict] = dict_obj.get('commands', {})
        for command_name, command_attrs in command_section.items():
            names: List[str] = always_list(command_attrs.pop('aliases', []))
            names.insert(0, command_name)
            param_section: Dict[str, dict] = command_attrs.pop('params', {})
            subcmd_section: Dict[str, dict] = command_attrs.pop('commands', {})
            command: "Params" = self.add_command(names, **command_attrs,
                                                 force=force)
            command.from_dict({"params": param_section,
                               "commands": subcmd_section}, show)

    def from_dict(self, dict_obj: dict, show: bool = True, force: bool = False):
        """Load parameters from python dict

        Args:
            dict_obj: A python dictionary to load parameters from
            show: The default show value for the parameters in the
                dictionary
            force: Whether to force adding params/commands
        """
        if 'params' not in dict_obj and 'commands' not in dict_obj:
            # express way
            # scan and create dict for all params
            all_params: Dict[str, dict] = {}
            for key, val in dict_obj.items():
                param_name: str = key
                if '.' in key:
                    param_name = key.split('.', 1)[0]
                all_params.setdefault(param_name, {})

            # scan for the attributes
            for key, val in dict_obj.items():
                if key in all_params:
                    # if it is a dict, and there is not other attributes
                    # assigned like "arg.desc", that means this dictionary
                    # contains all attributes
                    if (isinstance(val, dict) and
                            not any(ap_key.startswith(f"{key}.")
                                    for ap_key in dict_obj)):
                        all_params[key] = val
                    else:
                        all_params[key]['default'] = val
                elif '.' in key:
                    # Type: str, str
                    param_name, param_attr = key.split('.', 1)
                    all_params[param_name][param_attr] = val

            dict_obj = {"params": all_params}

        self._from_dict_with_sections(dict_obj, show, force)

    def from_arg(self, # pylint: disable=too-many-arguments
                 names: Union[str, List[str], 'ParamPath'],
                 desc: Union[str, List[str]] = "The configuration file.",
                 group: Optional[str] = None,
                 default: Optional[Union[str, Path]] = None,
                 show: bool = True,
                 force: bool = False,
                 args: Optional[List[str]] = None):
        """Load parameters from the file specified by naargument
        from command line

        This will load the parameters from the file given by the argument,
        ignoring other arguments from the command line. One can overwrite
        some of them afterwards, and do the parsing finally.

        Args:
            names: The names of the parameter or
                the parameter itself. If it is the parameter, other
                arguments are ignored
            desc: The description of the parameter
            group: The group of the parameter
            default: The default value of the file path
            show: Whether those parameters should show up in the help page
            force: Whether to force adding the parameters/commands
            args: The list of items to parse, otherwise
                parse sys.argv[1:]
        """
        if isinstance(names, (str, list)):
            param: "ParamPath" = self.add_param(
                names, desc=desc, group=group, type='path', default=default
            )
        else:
            param: "ParamPath" = names
            self._set_param(names)

        prefixed_names: List[str] = [param._prefix_name(name)
                                     for name in param.names]
        args: List[str] = sys.argv[1:] if args is None else args
        args_with_the_arg: List[str] = []
        for i, arg in enumerate(args):
            if arg in prefixed_names:
                args_with_the_arg.append(arg)
                if i < len(args) - 1:
                    args_with_the_arg.append(args[i+1])
                break
        filepath: Optional[Union[str, Path]] = (
            args_with_the_arg[1]
            if len(args_with_the_arg) > 1
            else param.value
        )
        if filepath:
            self.from_file(filepath, show=show, force=force)

    def __repr__(self):
        return "<Params(%s) @ %s>" % (','.join(self.names), id(self))

    def copy(self, deep=False) -> "Params":
        """Copy a Params object

        Args:
            deep: Whether to copy the parameters and commands deeply

        Returns:
            The copied params object
        """
        copied: "Params" = Params(
            names=self.names[:],
            desc=self.desc[:],
            prog=self.prog,
            help_keys=self.help_keys[:],
            help_cmds=self.help_cmds[:],
            help_on_void=self.help_on_void,
            help_callback=None,
            help_modifier=self.help_modifier,
            prefix=self.prefix,
            arbitrary=self.arbitrary,
            theme=self.theme,
            usage=self.usage and self.usage[:]
        )
        copied.assembler = self.assembler

        if not deep:
            copied.params = self.params.copy()
            copied.commands = self.commands.copy()

            copied.param_groups = self.param_groups.copy()
            copied.command_groups = self.command_groups.copy()
        else:
            copied.params = OrderedDiot()
            copied.commands = OrderedDiot()

            copied.param_groups = OrderedDiot()
            copied.command_groups = OrderedDiot()

            for param in self.params.values():
                if any(name in copied.params for name in param.names):
                    continue
                param_copy = param.copy()
                for name in param.names:
                    copied.params[name] = param_copy

            for command in self.commands.values():
                if any(name in copied.commands for name in command.names):
                    continue
                command_copy = command.copy(deep=deep)
                for name in command.names:
                    copied.commands[name] = command_copy

            for group, param_list in self.param_groups.items():
                self.param_groups[group] = [copied.get_param(param.names[0])
                                            for param in param_list]
            for group, cmd_list in self.command_groups.items():
                self.command_groups[group] = [copied.commands[cmd.names[0]]
                                              for cmd in cmd_list]

        return copied



"""Exceptions used in pyparam"""

class PyParamException(Exception):
    """Base exception for pyparam"""

class PyParamTypeError(PyParamException, TypeError):
    """When parameter type is not supported"""

class PyParamValueError(PyParamException, ValueError):
    """When parameter value is improper"""

class PyParamNameError(PyParamException, NameError):
    """Any errors related to parameter names"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyparam']

package_data = \
{'': ['*']}

install_requires = \
['diot', 'python-simpleconf', 'rich>=6,<7']

setup_kwargs = {
    'name': 'pyparam',
    'version': '0.4.2',
    'description': 'Powerful parameter processing.',
    'long_description': '# pyparam\n[![pypi][1]][2] [![pypi][10]][11] [![codacy quality][4]][6] [![codacy quality][7]][6] [![docs][12]][5] ![github action][3] ![pyver][8]\n\nPowerful parameter processing\n\n## Features\n- Command line argument parser (with subcommand support)\n- Rich type support, including `py`, `json`, `namespace`, etc.\n- Type overwriting for parameters from command line\n- Arbitrary parsing arguments from command line\n- Automatic help page assembling\n- Help page customization\n- Callbacks for option values\n- Parameter loading from configuration files\n\n## Installation\n```shell\npip install -U pyparam\n```\n\n## Documentation\n[https://pwwang.github.io/pyparam/][5]\n\n## Basic usage\n\n`example.py`\n\n```python\nfrom rich import print\nfrom pyparam import Params\n# program name, otherwise sys.argv[0]\nparams = Params(prog=\'pyparam\', desc="An example for {prog}")\n# adding parameters\nparams.add_param(\'i, int\', type=int,\n                 desc="An integer argument.")\nparams.add_param(\'float\', default=0.1, # type float implied\n                 desc="A float argument.")\nparams.add_param(\'str\', type=str,\n                 desc="A str argument.")\nparams.add_param(\'flag\', type=bool,\n                 desc="A flag argument.")\nparams.add_param(\'c,count\', type=\'count\',\n                 desc="A count argument.")\nparams.add_param(\'a\', type=\'auto\', type_frozen=False,\n                 desc="Value will be automatically casted.")\nparams.add_param(\'py\', type=\'py\',\n                 desc="Value will be evaluated by `ast.literal_eval`.")\nparams.add_param(\'json\', type=\'json\',\n                 desc="Value will be converted using `json.loads`.")\nparams.add_param(\'list\', type=\'list\',\n                 desc="Values will be accumulated.")\nparams.add_param(\'path\', type=\'path\', required=True,\n                 desc="Value will be casted into `pathlib.Path`.",\n                 callback=( # check if path exists\n                     lambda path: ValueError(\'File does not exist.\')\n                     if not path.exists() else path\n                 ))\nparams.add_param(\'choice\', type=\'choice\', default=\'medium\',\n                 choices=[\'small\', \'medium\', \'large\'],\n                 desc="One of {choices}.")\nparams.add_param(\'config.ncores\', default=1, # namespace config implied\n                 argname_shorten=False,\n                 desc=\'Number of cores to use.\')\n\nprint(vars(params.parse()))\n```\n\nTry it out:\n```sh\n$ python example.py\n```\n\n![help](./pyparam-help.png)\n\n```sh\n$ python example.py \\\n    -i2 \\\n    --float 0.5 \\\n    --str abc \\\n    -ccc \\\n    -a:int 1 \\\n    --py "{1,2,3}" \\\n    --json "{\\"a\\": 1}" \\\n    --list 1 2 3 \\\n    --choice large \\\n    --path . \\\n    --config.ncores 4\n```\n```python\n{\n    \'i\': 2,\n    \'int\': 2,\n    \'float\': 0.5,\n    \'str\': \'abc\',\n    \'flag\': False,\n    \'c\': 3,\n    \'count\': 3,\n    \'a\': 1,\n    \'py\': {1, 2, 3},\n    \'json\': {\'a\': 1},\n    \'list\': [1, 2, 3],\n    \'path\': PosixPath(\'.\'),\n    \'choice\': \'large\',\n    \'config\': Namespace(ncores=4)\n}\n```\n\nTry more features with:\n```sh\n$ python -m pyparam\n```\n\n## Shell completions\n\nHere is how the command completion in `fish` works:\n\n![pyparam-completions](./pyparam-completions.gif)\n\nCheck the [documentation][13], as well as the `__main__.py` to see how the completion works.\n\n\n[1]: https://img.shields.io/pypi/v/pyparam.svg?style=flat-square\n[2]: https://pypi.org/project/pyparam/\n[3]: https://img.shields.io/github/workflow/status/pwwang/pyparam/Build%20and%20Deploy?style=flat-square\n[4]: https://img.shields.io/codacy/grade/a34b1afaccf84019a6b138d40932d566.svg?style=flat-square\n[5]: https://pwwang.github.io/pyparam/\n[6]: https://app.codacy.com/project/pwwang/pyparam/dashboard\n[7]: https://img.shields.io/codacy/coverage/a34b1afaccf84019a6b138d40932d566.svg?style=flat-square\n[8]: https://img.shields.io/pypi/pyversions/pyparam.svg?style=flat-square\n[9]: https://raw.githubusercontent.com/pwwang/pyparam/master/docs/static/help.png\n[10]: https://img.shields.io/github/tag/pwwang/pyparam.svg?style=flat-square\n[11]: https://github.com/pwwang/pyparam\n[12]: https://img.shields.io/github/workflow/status/pwwang/pyparam/Build%20Docs?label=docs&style=flat-square\n[13]: https://pwwang.github.io/pyparam/shellCompletion/\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pwwang/pyparam',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Utilities for pyparam

Attributes:
    logger: The logger
"""
import logging
import ast
import json
import builtins
from typing import (
    Any,
    List,
    Union,
    Callable,
    Optional,
    Tuple
)
from functools import lru_cache
from argparse import Namespace as APNamespace
from pathlib import Path
from rich.logging import RichHandler
from rich.padding import Padding
from rich.syntax import Syntax
from .defaults import TYPE_NAMES
from .exceptions import PyParamTypeError

class Namespace(APNamespace):
    """Subclass of `argparse.Namespace`

    We have enabled `__getitem__`, `__setitem__`, `__len__` and `__contains__`.
    So that you can do:

    ```python
    ns = Namespace()
    ns['a'] = 1  # same as ns.a = 1
    ns['a'] == 1 # same as ns.a == 1
    len(ns) == 1
    'a' in ns
    ```

    Attributes:
        __command__: The command name if matched.
    """
    __command__: Optional[str] = None

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name)

    def __setitem__(self, name: str, value: Any) -> None:
        setattr(self, name, value)

    def __len__(self) -> int:
        return len(vars(self))

    def __contains__(self, name: str) -> bool:
        return name in vars(self)

    def __or__(self, other):
        # copy myself
        myself: Namespace = self.__class__()
        myself |= self
        myself |= other
        return myself

    def __ior__(self, other):
        if not isinstance(other, dict):
            other = vars(other)
        for key, value in other.items():
            self[key] = value
        return self

class Codeblock:
    """A code block, will be rendered as rich.syntax.Syntax"""

    @classmethod
    def scan_texts(
            cls,
            texts: List[str],
            check_default: bool = False
    ) -> List[Union[str, 'Codeblock']]:
        """Scan multiple texts for code blocks

        Args:
            cls (Codeblock class): The class
            texts: a list of texts
            check_default: Check if there is default in maybe_codeblock.
                Defaults should not be scanned as code blocks

        Returns:
            mixed text and code blocks
        """

        ret: List[Union[str, 'Codeblock']] = []
        ret_extend: Callable = ret.extend
        codeblock: Optional['Codeblock'] = None
        # Type: Union[str, Codeblock]
        for text in texts:
            if not codeblock:
                if not text:
                    ret.append(text)
                    continue
                # Type: List[Union[str, Codeblock]], Optional[Codeblock]
                scanned, codeblock = cls.scan(text, check_default=check_default)
                ret_extend(scanned)
                continue
            # if we hit an unclosed codeblock
            lines: List[str] = text.splitlines()
            # Type: int, str
            for i, line in enumerate(lines):
                if codeblock.is_end(line):
                    # Type: List[Union[str, Codeblock]], Optional[Codeblock]
                    scanned, codeblock = cls.scan(
                        '\n'.join(
                            lines[(i if codeblock.opentag == '>>>' else i+1):]
                        ),
                        check_default=check_default
                    )
                    ret_extend(scanned)
                    break
            else:
                codeblock.add_code(text)
        return ret

    @classmethod
    def scan(
            cls,
            maybe_codeblock: str,
            check_default: bool = False
    ) -> Tuple[List[Union[str, 'Codeblock']], Optional['Codeblock']]:
        """Scan and try to create codeblock objects from maybe_codeblock

        Args:
            cls (Codeblock class): The class
            maybe_codeblock: Maybe a code block start
                It can be a text block, we have to scan if it has code blocks
                inside.
            check_default: Check if there is default in maybe_codeblock.
                Defaults should not be scanned as code blocks

        Returns:
            mixed text and unclosed code blocks
        """
        sep: Optional[str] = (
            'Default:' if 'Default:' in maybe_codeblock
            else 'DEFAULT:' if 'DEFAULT:' in maybe_codeblock
            else None
        )

        default_to_append: Optional[str] = None
        default_in_newline = False
        if check_default and sep:
            parts: List[str] = maybe_codeblock.split(sep, 1)
            default_to_append = sep + parts[1]
            default_in_newline = parts[0].endswith('\n')
            lines: List[str] = parts[0].splitlines()
        else:
            lines: List[str] = maybe_codeblock.splitlines()

        ret: List[Union[str, 'Codeblock']] = []
        ret_append: Callable = ret.append
        codeblock: Optional['Codeblock'] = None
        # Type: str
        for line in lines:
            if not codeblock:
                line_lstripped: str = line.lstrip()
                if line_lstripped.startswith('>>>'):
                    codeblock: 'Codeblock' = cls(
                        '>>>',
                        'pycon',
                        len(line) - len(line_lstripped),
                        [line_lstripped]
                    )
                    ret_append(codeblock)
                elif line_lstripped.startswith('```'):
                    codeblock: 'Codeblock' = cls(
                        line_lstripped[
                            :(len(line_lstripped) -
                              len(line_lstripped.lstrip('`')))
                        ],
                        line_lstripped.lstrip('`').strip() or 'text',
                        len(line) - len(line_lstripped)
                    )
                    ret_append(codeblock)
                else:
                    ret_append(line)
            elif codeblock.is_end(line):
                if codeblock.opentag == '>>>':
                    ret.append(line)
                codeblock = None
            else:
                codeblock.add_code(line)

        if default_to_append:
            # if codeblock (>>>) is not closed.
            # but we have default so it actually closes
            if codeblock and codeblock.opentag == '>>>':
                codeblock = None
            if not ret or isinstance(ret[-1], Codeblock) or default_in_newline:
                ret.append(default_to_append)
            else:
                ret[-1] += default_to_append
        return ret, codeblock

    def __init__(self,
                 opentag: str,
                 lang: str,
                 indent: int,
                 codes: Optional[List[str]] = None) -> None:
        """Constructor

        Args:
            opentag: The opentag for the code block.
                One of '>>>', '```<lang>', '````<lang>', ...
            lang: The language name
            indent: The indentation level
            codes: The lines of code
        """
        self.opentag: str = opentag
        self.lang: str = lang
        self.indent: int = indent
        self.codes: List[str] = codes or []

    def __repr__(self) -> str:
        return (f"<Codeblock (tag={self.opentag}, lang={self.lang}, "
                f"codes={self.codes[:1]} ...)")

    def add_code(self, code: str) -> None:
        """Add code to code block

        Args:
            code: code to add
                It can be multiple lines, each of which will be dedented
        """
        # Type: str
        for line in code.splitlines():
            self.codes.append(line[self.indent:])

    def is_end(self, line: str) -> bool:
        """Tell if the line is the end of the code block

        Args:
            line: line to check

        Returns:
            True if it is the end otherwise False
        """
        if self.opentag == '>>>' and not line[self.indent:].startswith('>>>'):
            return True
        if '`' in self.opentag and line[self.indent:].rstrip() == self.opentag:
            return True
        return False

    def render(self) -> Padding:
        """Render the code block to a rich.syntax.Syntax

        Returns:
            A padding of rich.syntax.Syntax
        """
        return Padding(
            Syntax('\n'.join(self.codes), self.lang),
            (0, 0, 0, self.indent)
        )

def always_list(str_or_list: Union[str, List[str]],
                strip: bool = True,
                split: Union[str, bool] = ',') -> List[str]:
    """Convert a string (comma separated) or a list to a list

    Args:
        str_or_list: string or list
        strip: whether to strip the elements in result list
        split: Delimiter for split or False to not split

    Return:
        list: list of strings
    """
    if isinstance(str_or_list, (list, tuple)):
        return list(str_or_list)
    if split:
        return [elem.strip() if strip else elem
                for elem in str_or_list.split(split)]
    return [str_or_list]

def parse_type(typestr: str) -> List[Optional[str]]:
    """Parse the type string

    Examples:
        >>> parse_type(None)    # None, None
        >>> parse_type("array") # list, None
        >>> parse_type("a:i")   # list, int
        >>> parse_type("j")     # json, None
        >>> parse_type("list")  # list, None

    Args:
        typestr: string of type to parse

    Returns:
        Main type and subtype

    Raises:
        PyParamTypeError: When a type cannot be parsed
    """
    if typestr is None:
        return [None, None]

    parts: List[str] = typestr.split(':', 1)
    # Type: int, str
    for i, part in enumerate(parts):
        if part not in TYPE_NAMES:
            raise PyParamTypeError("Unknown type: %s" % typestr)
        parts[i] = TYPE_NAMES[part]

    parts.append(None)
    return parts[:2]

@lru_cache()
def parse_potential_argument(
        arg: str,
        prefix: str,
        allow_attached: bool = False
) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """Parse a potential argument with given prefix

    Examples:
        >>> # prefix == 'auto
        >>> parse_potential_argument("-a")     # a, None, None
        >>> parse_potential_argument("--arg")  # arg, None, None
        >>> parse_potential_argument("--a")    # None, None, a
        >>> parse_potential_argument("-abc")   # None, None, -abc
        >>> parse_potential_argument("-abc", allow_attached=True)
        >>> # -a, None, bc

    Args:
        arg: a potential argument. Such as:
            -a, --def, -b=1, --abc=value, -b1 (for argument -b with value 1)
            with types:
            -a:int --def:list -b:str=1 --abs:str=value -b:bool
            It is usually one element of the sys.argv
        prefix: The prefix for the argument names
        allow_attached: Whether to detect item like '-b1' for argument
            '-b' with value '1' or the entire one is parsed as argument '-b1'

    Returns:
        The argument name, type and value
            When arg cannot be parsed as an argument, argument name and type
            will both be None. arg will be returned as argument value.
    """
    if not arg.startswith('-' if prefix == 'auto' else prefix):
        return None, None, arg

    # fill a tuple to length of 2 with None
    fill2_none: Callable = lambda alist: (
        (alist[0], None) if len(alist) == 1 or not alist[1] else alist[:2]
    )

    # Type: str, Optional[str]
    item_nametype, item_value = fill2_none(arg.split('=', 1))
    # Type: str, Optional[str]
    item_name, item_type = fill2_none(item_nametype.split(':', 1))

    # detach the value for -b1
    if allow_attached:
        single_prefix: str = '-' if prefix == 'auto' else prefix
        len_spref: int = len(single_prefix)
        if (item_type is None and item_value is None and (
                item_name.startswith(single_prefix) and
                item_name[len_spref:len_spref+1] != single_prefix
        )):
            # Type: str, str
            item_name, item_value = (
                single_prefix + item_name[len_spref:len_spref+1],
                item_name[len_spref+1:]
            )

    # remove prefix in item_name
    if prefix == 'auto':
        item_name_first: str = item_name.split('.')[0]
        prefix = ('-' if len(item_name_first) <= 2
                  else '--' if len(item_name_first) >= 4
                  else None)

    if prefix is not None and item_name.startswith(prefix):
        item_name = item_name[len(prefix):]
    else:
        return None, None, arg

    # Type: str, Optional[str]
    item_type, item_subtype = parse_type(item_type)
    item_type = f"{item_type}:{item_subtype}" if item_subtype else item_type

    return item_name, item_type, item_value

def type_from_value(value: Any) -> str:
    """Detect parameter type from a value

    Args:
        value: The value

    Returns:
        The name of the type

    Raises:
        PyParamTypeError: When we have list as subtype.
            For example, when value is `[[1]]`
    """
    typename: str = type(value).__name__
    if typename in ('int', 'float', 'str', 'bool'):
        return typename
    if isinstance(value, list):
        if not value:
            return 'list'
        type0: str = type_from_value(value[0])
        if 'list' in type0:
            raise PyParamTypeError("Cannot have 'list' as subtype.")
        return (f'list:{type0}'
                if all(type_from_value(item) == type0 for item in value[1:])
                else 'list')
    if isinstance(value, dict):
        return 'json'
    if isinstance(value, Path):
        return 'path'
    return 'auto'

def _cast_auto(value: Any) -> Any:
    """Cast value automatically

    Args:
        value: value to cast

    Returns:
        value casted
    """
    # pylint: disable=too-many-return-statements
    if value in ("True", "TRUE", "true") or value is True:
        return True
    if value in ("False", "FALSE", "false") or value is False:
        return False

    try:
        return int(value)
    except (TypeError, ValueError):
        pass

    try:
        return float(value)
    except (TypeError, ValueError):
        pass

    try:
        return json.loads(value)
    except (TypeError, ValueError, json.JSONDecodeError):
        pass

    return value

def cast_to(value: Any, to_type: str) -> Any:
    """Cast a value to a given type

    Args:
        value: value to cast
        to_type: type to cast

    Returns:
        casted value

    Raises:
        PyParamTypeError: if value is not able to be casted
    """
    try:
        if to_type in ('int', 'float', 'str'):
            return getattr(builtins, to_type)(value)
        if to_type == 'bool':
            if value in ('true', 'TRUE', 'True', '1', 1, True):
                return True
            if value in ('false', 'FALSE', 'False', '0', 0, False):
                return False
            raise PyParamTypeError(
                'Expecting one of [true, TRUE, True, 1, false, FALSE, False, 0]'
            )

        if to_type in ('path', 'py', 'json'):
            return {
                'path': Path,
                'py': ast.literal_eval,
                'json': json.loads
            }[to_type](str(value))
        if to_type in (None, 'auto'):
            return _cast_auto(value)
    except (TypeError, ValueError, json.JSONDecodeError) as cast_exc:
        raise PyParamTypeError(
            f"Cannot cast {value} to {to_type}: {cast_exc}"
        ) from cast_exc
    raise PyParamTypeError(f"Cannot cast {value} to {to_type}")

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(RichHandler(show_time=False, show_path=False))



"""Help assembler for pyparam

Attributes:
    THEMES: The theme for the help page.
"""
import re
import textwrap
from typing import (
    Dict,
    List,
    Optional,
    Type,
    Union,
    Callable,
    Tuple
)
from diot import OrderedDiot, Diot
from rich import box#, print
from rich.table import Table
from rich.padding import Padding
from rich.console import Console, RenderResult, RenderGroup
from rich.theme import Theme
from rich.text import Text
from rich.highlighter import RegexHighlighter
from . import defaults
from .utils import Codeblock

THEMES: Dict[str, Theme] = dict(
    default=Theme(dict(
        title="bold cyan",
        prog="bold green",
        default="magenta",
        optname="bright_green",
        opttype="blue italic",
        opttype_frozen="blue"
    )),
    synthware=Theme(dict(
        title="bold magenta",
        prog="bold yellow",
        default="cyan",
        optname="bright_yellow",
        opttype="bright_red italic",
        opttype_frozen="bright_red"
    ))
)

# pylint: disable=too-few-public-methods
class ProgHighlighter(RegexHighlighter):
    """Apply style to anything that looks like a program name.

    Args:
        prog: The program name
    """

    def __init__(self, prog: str):
        super().__init__()
        prog = re.escape(prog)
        self.highlights = [rf"(?P<prog>\b{prog}\b)"]

class OptnameHighlighter(RegexHighlighter):
    """Apply style to anything that looks like a option name.

    Highlight `-b` and `--box` in `-b, --box <INT>`, and all in commands:
    `i, install`
    """

    highlights: List[str] = [r"(?P<optname>[^\[<][^,\s]+)"]

class OpttypeHighlighter(RegexHighlighter):
    """Apply style to anything that looks like a option type."""

    highlights: List[str] = [
        r"(?P<opttype_frozen>[\[\<][A-Z:]+[\]\>])$",
        r"(?P<opttype>[\[\<][a-z:]+[\]\>])$",
    ]

class DefaultHighlighter(RegexHighlighter):
    """Apply style to anything that looks like default value in option desc."""

    highlights: List[str] = [r"(?P<default>D(?:efault|EFAULT):.+$)"]

class HelpSection(list):
    """Base class for all help sections."""

    def _highlight( # pylint: disable=no-self-use
            self,
            string: str,
            highlighters: Optional[List[Type[RegexHighlighter]]] = None
    ) -> Union[Text, str]:
        """Highlight the string using given highlighters"""
        if not highlighters:
            return string
        if not isinstance(highlighters, (tuple, list)):
            highlighters = [highlighters]
        for highlighter in highlighters:
            string = highlighter(string)

        return string

    def __rich_console__(self, console: Console, _) -> RenderResult:
        """Implement API from rich to print the help page"""
        scanned = Codeblock.scan_texts(self)
        for item in scanned:
            if isinstance(item, Codeblock):
                yield Padding(item.render(), (0, 0, 0,
                                              defaults.HELP_SECTION_INDENT))
            else:
                yield Padding(
                    self._highlight(item, console.meta.highlighters.prog),
                    (0, 0, 0, defaults.HELP_SECTION_INDENT)
                )

class HelpSectionPlain(HelpSection):
    """Plain text section"""

class HelpSectionUsage(HelpSectionPlain):
    """Usage section in help"""

    def _wrap_usage(self,
                    usage: str,
                    prog: str,
                    *highlighters) -> Union[Text, str]:
        """Wrap usage line"""
        for line in textwrap.wrap(
                usage,
                width=defaults.CONSOLE_WIDTH,
                initial_indent=' ' * defaults.HELP_SECTION_INDENT,
                subsequent_indent=' ' * (defaults.HELP_SECTION_INDENT +
                                         len(prog) + 1),
                break_long_words=False,
                break_on_hyphens=False
        ):
            yield self._highlight(line.replace('*', ' '), highlighters)

    def __rich_console__(self, console: Console, _) -> RenderResult:
        """Implement API from rich to print the help page"""
        for line in self:
            usages = self._wrap_usage(line,
                                      console.meta.prog,
                                      console.meta.highlighters.prog)

            yield RenderGroup(*usages)

class HelpSectionOption(HelpSection):
    """Options section in help"""

    def _wrap_opts(self, opts: List[str], *highlighters) -> Union[Text, str]:
        """Wrap the option names and types"""
        for opt in opts:
            for line in textwrap.wrap(
                    opt,
                    width=defaults.HELP_OPTION_WIDTH,
                    initial_indent=' ' * defaults.HELP_SECTION_INDENT,
                    subsequent_indent=' ' * (
                        defaults.HELP_SECTION_INDENT +
                        len(opt.split(",")[0]) + 2
                    ),
                    break_long_words=False,
                    break_on_hyphens=False
            ):
                yield self._highlight(line.replace('*', ' '), highlighters)

    def _wrap_descs(
            self,
            descs: List[str],
            default_highlighter: DefaultHighlighter
    ) -> Union[Text, str]:
        """wrap option descriptions.

        Highlight default value, inline code and code blocks

        Code blocks could be either markdown style:
        ```python
        print('Hello world!')
        ```

        Or a python console style:
        >>> print('Hello world!')
        """
        hillight_inline_code: Callable = (
            lambda text: re.sub(r'(`+)(.+?)\1', r'[code]\2[/code]', text)
        )

        descs = Codeblock.scan_texts(descs, check_default=True)
        def wrap_normal(text):
            for line in textwrap.wrap(
                    text,
                    drop_whitespace=True,
                    width=(defaults.CONSOLE_WIDTH -
                           defaults.HELP_OPTION_WIDTH - 2)
            ):
                yield self._highlight(hillight_inline_code(line))

        for desc in descs: # pylint: disable=too-many-nested-blocks
            if isinstance(desc, Codeblock):
                yield desc.render()

            else:
                sep: Optional[str] = (
                    'Default:' if 'Default:' in desc
                    else 'DEFAULT:' if 'DEFAULT:' in desc
                    else None
                )

                if sep:
                    parts: List[str] = desc.split(sep, 1)
                    # if default is multiline, put it in new line
                    if "\n" in parts[1]:
                        yield from wrap_normal(parts[0])
                        for i, line in enumerate(
                                parts[1].lstrip().splitlines()
                        ):
                            if i == 0:
                                yield Text(sep + ' ' + line, style="default")
                            else:
                                yield Text(" " * (len(sep) + 1) + line,
                                           style="default")
                    else:
                        # use * to connect to avoid default to be wrapped
                        parts[0] += sep + '*' * len(parts[1])

                        # wrap default
                        for line in textwrap.wrap(
                                parts[0],
                                width=(defaults.CONSOLE_WIDTH -
                                       defaults.HELP_OPTION_WIDTH - 2),
                                break_long_words=False,
                                break_on_hyphens=False
                        ):
                            yield self._highlight(
                                Text.from_markup(
                                    hillight_inline_code(line).replace(
                                        sep + '*' * len(parts[1]),
                                        sep + parts[1].replace('[', r'\[')
                                    )
                                ),
                                default_highlighter
                            )
                elif not desc:
                    yield ''
                else:
                    yield from wrap_normal(desc)

    def __rich_console__(self, console: Console, _) -> RenderResult:
        """Implement API from rich to print the help page"""
        table = Table(width=defaults.CONSOLE_WIDTH,
                      show_header=False,
                      show_lines=False,
                      show_edge=False,
                      box=box.SIMPLE,
                      expand=True,
                      pad_edge=False,
                      padding=(0, 0, 0, 0))
        table.add_column(width=defaults.HELP_OPTION_WIDTH)
        table.add_column(width=1)
        table.add_column(width=defaults.CONSOLE_WIDTH -
                         defaults.HELP_OPTION_WIDTH - 1)
        for param_opts, param_descs in self:
            table.add_row(
                RenderGroup(*self._wrap_opts(
                    param_opts,
                    console.meta.highlighters.optname,
                    console.meta.highlighters.opttype
                )),
                Text('-', justify='left'),
                RenderGroup(*self._wrap_descs(
                    param_descs or [],
                    console.meta.highlighters.default
                ))
            )
        yield table

class HelpAssembler:
    """Assemble a help page

    Args:
        prog: The name of the program
        theme: The theme for the help page

    Attributes:
        console: The console to print the help page
        callback: The callback to modify the help page
    """

    def __init__(self,
                 prog: str,
                 theme: Union[str, Theme],
                 callback: Optional[Callable]):
        """Constructor"""
        theme = (theme if isinstance(theme, Theme)
                 else THEMES.get(theme, 'default'))

        self.console: Console = Console(
            theme=theme, width=defaults.CONSOLE_WIDTH, tab_size=4
        )
        self.callback: Optional[Callable] = callback
        self._assembled: Optional[List[RenderResult]] = None

        self.console.meta = Diot()
        self.console.meta.prog = prog
        self.console.meta.highlighters = Diot()
        self.console.meta.highlighters.prog = ProgHighlighter(prog)
        self.console.meta.highlighters.optname = OptnameHighlighter()
        self.console.meta.highlighters.opttype = OpttypeHighlighter()
        self.console.meta.highlighters.default = DefaultHighlighter()

    def _assemble_description(self, # pylint: disable=no-self-use
                              params: "Params") -> Optional[HelpSectionPlain]:
        """Assemble the description section"""
        if not params.desc:
            return None

        return HelpSectionPlain(
            desc.format(prog=params.prog) for desc in params.desc
        )

    def _assemble_usage(self, params: "Params") -> HelpSectionUsage:
        """Assemble the usage section"""
        # pylint: disable=no-self-use
        if not params.usage:
            # default usage
            # gather required Arguments
            usage: List[str] = ['{prog}']
            has_optional = False


            for group in params.param_groups.values():
                for param in group:
                    if param.required and param.show:
                        usage.append(param.usagestr())
                    elif param.show:
                        has_optional = True
            if has_optional:
                usage.append('[OPTIONS]')

            if params.commands:
                usage.append('COMMAND [OPTIONS]')

            params.usage = [" ".join(usage)]

        return HelpSectionUsage(
            usage.format(prog=params.prog) for usage in params.usage
        )

    def _assemble_param_groups( # pylint: disable=no-self-use
            self,
            params: "Params"
    ) -> Tuple[str, HelpSectionOption]:
        """Assemble the parameter groups"""

        for group, param_list in params.param_groups.items():
            if all(not param.show for param in param_list):
                continue

            yield group, HelpSectionOption(
                ([param.optstr()],
                 param.desc_with_default)
                for param in param_list if param.show
            )

    def _assemble_command_groups( # pylint: disable=no-self-use
            self,
            params: "Params"
    ) -> Tuple[str, HelpSectionOption]:
        """Assemble the command groups"""
        # command groups
        for group, cmd_list in params.command_groups.items():

            yield group, HelpSectionOption(
                ([command.namestr()],
                 command.desc)
                for command in cmd_list
            )

    def assemble(self, params: "Params") -> None:
        """Assemble the help page

        Args:
            params: The params object
        """
        self._assembled = []

        assembled: OrderedDiot = OrderedDiot()

        assembled_description: Optional[HelpSectionPlain] = (
            self._assemble_description(params)
        )
        if assembled_description:
            assembled.DESCRIPTION = assembled_description

        assembled_usage: HelpSectionPlain = self._assemble_usage(params)
        assembled.USAGE = assembled_usage

        for group, section in self._assemble_param_groups(params):
            assembled[group] = section

        for group, section in self._assemble_command_groups(params):
            assembled[group] = section

        if callable(self.callback):
            self.callback(assembled)

        for title, section in assembled.items():
            self._assembled.append(Text(end="\n"))
            self._assembled.append(Text(title + ':', style="title",
                                        justify="left"))
            self._assembled.append(section)

    def printout(self) -> None:
        """Print the help page"""
        self.console.print(*self._assembled)
